# AUTOGENERATED! DO NOT EDIT! File to edit: notebooks/FDIC_Banks.ipynb (unless otherwise specified).

__all__ = ['banks', 'banks', 'banks', 'banks', 'banks']

# Cell
banks = gpd.read_file('BCityBanks_2019_CSACity.shp')
banks = banks.fillna('Baltimore City')
banks['count'] = 1
banks= banks[['CSA','count']]
banks = banks.groupby('CSA').sum(numeric_only=True)
# Make sure ALL csas and BaltimoreCity are included. among other things
banks = csa[ ['CSA2010','tpop10'] ].merge( banks, left_on='CSA2010', right_on='CSA', how='outer' )
# Update the baltimore CSA.
banks.at[55,'count'] = banks['count'].sum()
banks.head(3)

# Cell
# Create the Indicator
banks['banks19'] = banks['count'] * 1000 / banks['tpop10']



# AUTOGENERATED! DO NOT EDIT! File to edit: notebooks/00_core.ipynb (unless otherwise specified).

__all__ = ['core_test_fn', 'check_labels']

# Cell
def core_test_fn(txt):
  return txt + txt + txt + txt

# Cell
def check_labels(ds):
  original = ds.copy()
  original.rename(columns={ 'CSA':'CSA2010', 'BaltCity':'InBaltimore'}, inplace=True)
  df = original[ original['CSA2010'].notnull() | original['InBaltimore'].notnull()  ]
  print( 'Total # Rows: ', original.shape[0] ) # rows, columns
  print( '# Before | After')
  print( '# Where BCity.isnull/notnull: ', original.InBaltimore.isnull().sum(), '|', original.InBaltimore.notnull().sum() );
  print( '# where CSA2010.isnull/notnull: ', original.CSA2010.isnull().sum(), '|', original.CSA2010.notnull().sum() );
  print( '# Where CSA and/or Baltimore lbl Exists: ', df.shape[0])
  df.CSA2010 = df.CSA2010.fillna('Baltimore City')
  return df


# AUTOGENERATED! DO NOT EDIT! File to edit: notebooks/DHR_SNAP_TANF_Create.ipynb (unless otherwise specified).

__all__ = ['tanf', 'snap']

# Cell
def tanf(df, csa, yr):

  # Create the Numerator
  tanf = df.copy()
  tanf['count'] = 1
  tanf = tanf.groupby('CSA2010').sum(numeric_only=True)

  # Make sure ALL csas and BaltimoreCity are included and sorted.
  tanf = csa.merge( tanf, left_on='CSA2010', right_on='CSA2010', how='outer' )
  # Baltimoire may have records not in the CSA (not actually the case right now but..)
  tanf.at[55,'count']=tanf['count'].sum()
  # Perform the calculation
  tanf['106-tanf'+year] = tanf['count'] / tanf['FamHH_2010'] * 100

  compareYears = gpd.read_file("https://services1.arcgis.com/mVFRs7NF4iFitgbY/ArcGIS/rest/services/Tanf/FeatureServer/0/query?where=1%3D1&outFields=*&returnGeometry=true&f=pgeojson");
  prevYear = 'tanf'+ str( int(year) - 1 )
  if prevYear in compareYears.columns:
    tanf = tanf.merge( compareYears[['CSA2010', prevYear]], left_on='CSA2010', right_on='CSA2010', how='outer' )
    tanf['change'] = tanf['106-tanf'+year] - tanf[ prevYear ]
    tanf['percentChange'] = tanf['change' ] / tanf[ prevYear ] * 100
    tanf['change'] = tanf['change'].apply(lambda x: "{:.2f}".format(x) )
  print( 'Records Matching Query: ', tanf.size / len(tanf.columns) )
  return tanf

# Cell
def snap(df, csa, yr):

  # Create the Numerator
  snap = df.copy()
  snap['count'] = 1
  snap = snap.groupby('CSA2010').sum(numeric_only=True)

  # Make sure ALL csas and BaltimoreCity are included and sorted.
  snap = csa.merge( snap, left_on='CSA2010', right_on='CSA2010', how='outer' )
  # Baltimoire may have records not in the CSA (not actually the case right now but..)
  snap.at[55,'count']=snap['count'].sum()
  # Perform the calculation
  snap['???-snap'+year] = snap['count']

  compareYears = gpd.read_file("https://services1.arcgis.com/mVFRs7NF4iFitgbY/ArcGIS/rest/services/Snap/FeatureServer/0/query?where=1%3D1&outFields=*&returnGeometry=true&f=pgeojson");
  prevYear = 'snap'+ str( int(year) - 1 )
  if prevYear in compareYears.columns:
    snap = snap.merge( compareYears[['CSA2010', prevYear]], left_on='CSA2010', right_on='CSA2010', how='outer' )
    snap['change'] = snap['???-snap'+year] - snap[ prevYear ]
    snap['percentChange'] = snap['change' ] / snap[ prevYear ] * 100
    snap['change'] = snap['change'].apply(lambda x: "{:.2f}".format(x) )
  print( 'Records Matching Query: ', snap.size / len(snap.columns) )
  return snap


# AUTOGENERATED BY NBDEV! DO NOT EDIT!

__all__ = ["index", "modules", "custom_doc_links", "git_url"]

index = {"core_test_fn": "00_core.ipynb",
         "check_labels": "00_core.ipynb",
         "HEADER": "01_Close Crawl.ipynb",
         "URL": "01_Close Crawl.ipynb",
         "CASE_PAT": "01_Close Crawl.ipynb",
         "CASE_TYPES": "01_Close Crawl.ipynb",
         "HTML_DIR": "01_Close Crawl.ipynb",
         "HTML_FILE": "01_Close Crawl.ipynb",
         "CHECKPOINT": "01_Close Crawl.ipynb",
         "NO_CASE": "01_Close Crawl.ipynb",
         "FEATURES": "01_Close Crawl.ipynb",
         "FIELDS": "01_Close Crawl.ipynb",
         "INTERNAL_FIELDS": "01_Close Crawl.ipynb",
         "Session": "01_Close Crawl.ipynb",
         "Spider": "01_Close Crawl.ipynb",
         "scrape": "01_Close Crawl.ipynb",
         "filter_addr": "01_Close Crawl.ipynb",
         "PUNCTUATION": "01_Close Crawl.ipynb",
         "street_address": "01_Close Crawl.ipynb",
         "TITLE_SPLIT_PAT": "01_Close Crawl.ipynb",
         "ZIP_STR": "01_Close Crawl.ipynb",
         "ZIP_PAT": "01_Close Crawl.ipynb",
         "MONEY_STR": "01_Close Crawl.ipynb",
         "MONEY_PAT": "01_Close Crawl.ipynb",
         "NULL_ADDR": "01_Close Crawl.ipynb",
         "STRIP_ADDR": "01_Close Crawl.ipynb",
         "Miner": "01_Close Crawl.ipynb",
         "temp_output": "01_Close Crawl.ipynb",
         "file_array": "01_Close Crawl.ipynb",
         "top5": "01_Close Crawl.ipynb",
         "Cleaner": "01_Close Crawl.ipynb",
         "output": "01_Close Crawl.ipynb",
         "df_obj": "01_Close Crawl.ipynb",
         "close_crawl": "01_Close Crawl.ipynb",
         "args": "01_Close Crawl.ipynb",
         "runTidyAddr": "02_TidyAddr.ipynb",
         "affordm": "04_Create_Acs_Indicators_Original.ipynb",
         "affordr": "04_Create_Acs_Indicators_Original.ipynb",
         "age5": "05_Create_Acs_Indicators.ipynb",
         "age18": "05_Create_Acs_Indicators.ipynb",
         "age24": "04_Create_Acs_Indicators_Original.ipynb",
         "age64": "04_Create_Acs_Indicators_Original.ipynb",
         "age65": "05_Create_Acs_Indicators.ipynb",
         "bahigher": "04_Create_Acs_Indicators_Original.ipynb",
         "carpool": "04_Create_Acs_Indicators_Original.ipynb",
         "drvalone": "04_Create_Acs_Indicators_Original.ipynb",
         "elheat": "04_Create_Acs_Indicators_Original.ipynb",
         "empl": "04_Create_Acs_Indicators_Original.ipynb",
         "fam": "04_Create_Acs_Indicators_Original.ipynb",
         "female": "04_Create_Acs_Indicators_Original.ipynb",
         "femhhs": "04_Create_Acs_Indicators_Original.ipynb",
         "heatgas": "04_Create_Acs_Indicators_Original.ipynb",
         "hh25inc": "05_Create_Acs_Indicators.ipynb",
         "hh40inc": "04_Create_Acs_Indicators_Original.ipynb",
         "hh60inc": "04_Create_Acs_Indicators_Original.ipynb",
         "hh75inc": "04_Create_Acs_Indicators_Original.ipynb",
         "hhchpov": "04_Create_Acs_Indicators_Original.ipynb",
         "hhm75": "04_Create_Acs_Indicators_Original.ipynb",
         "hhpov": "04_Create_Acs_Indicators_Original.ipynb",
         "hhs": "04_Create_Acs_Indicators_Original.ipynb",
         "hsdipl": "04_Create_Acs_Indicators_Original.ipynb",
         "lesshs": "04_Create_Acs_Indicators_Original.ipynb",
         "male": "04_Create_Acs_Indicators_Original.ipynb",
         "mhhi": "05_Create_Acs_Indicators.ipynb",
         "nilf": "04_Create_Acs_Indicators_Original.ipynb",
         "nohhint": "05_Create_Acs_Indicators.ipynb",
         "novhcl": "05_Create_Acs_Indicators.ipynb",
         "othrcom": "04_Create_Acs_Indicators_Original.ipynb",
         "p2more": "04_Create_Acs_Indicators_Original.ipynb",
         "paa": "05_Create_Acs_Indicators.ipynb",
         "pasi": "04_Create_Acs_Indicators_Original.ipynb",
         "phisp": "04_Create_Acs_Indicators_Original.ipynb",
         "ppac": "04_Create_Acs_Indicators_Original.ipynb",
         "pubtran": "04_Create_Acs_Indicators_Original.ipynb",
         "pwhite": "05_Create_Acs_Indicators.ipynb",
         "racdiv": "04_Create_Acs_Indicators_Original.ipynb",
         "retrieve_acs_data": "90_ACS_Explore_and_Download.ipynb",
         "sclemp": "04_Create_Acs_Indicators_Original.ipynb",
         "tpop": "04_Create_Acs_Indicators_Original.ipynb",
         "trav14": "04_Create_Acs_Indicators_Original.ipynb",
         "trav29": "04_Create_Acs_Indicators_Original.ipynb",
         "trav44": "05_Create_Acs_Indicators.ipynb",
         "trav45": "05_Create_Acs_Indicators.ipynb",
         "unempl": "04_Create_Acs_Indicators_Original.ipynb",
         "unempr": "04_Create_Acs_Indicators_Original.ipynb",
         "walked": "04_Create_Acs_Indicators_Original.ipynb",
         "getColName": "05_Create_Acs_Indicators.ipynb",
         "getColByName": "05_Create_Acs_Indicators.ipynb",
         "addKey": "05_Create_Acs_Indicators.ipynb",
         "nullIfEqual": "05_Create_Acs_Indicators.ipynb",
         "sumInts": "05_Create_Acs_Indicators.ipynb",
         "hisp": "05_Create_Acs_Indicators.ipynb",
         "createAcsIndicator": "05_Create_Acs_Indicators.ipynb",
         "InteractiveShell.ast_node_interactivity": "90_ACS_Explore_and_Download.ipynb",
         "eattend": "BCPSS_Schools.ipynb",
         "mattend": "BCPSS_Schools.ipynb",
         "hsattend": "BCPSS_Schools.ipynb",
         "aastud": "BCPSS_Schools.ipynb",
         "wstud": "BCPSS_Schools.ipynb",
         "hstud": "BCPSS_Schools.ipynb",
         "abse": "BCPSS_Schools.ipynb",
         "absmd": "BCPSS_Schools.ipynb",
         "abshd": "BCPSS_Schools.ipynb",
         "drop": "BCPSS_Schools.ipynb",
         "compl": "BCPSS_Schools.ipynb",
         "sclsw": "BCPSS_Schools.ipynb",
         "eenrol": "BCPSS_Schools.ipynb",
         "menrol": "BCPSS_Schools.ipynb",
         "hsenrol": "BCPSS_Schools.ipynb",
         "taxlien": "Bidbalt_TaxSale_Taxlien.ipynb",
         "dirtyst": "CitiStat_Dirtyst_Clogged.ipynb",
         "clogged": "CitiStat_Dirtyst_Clogged.ipynb",
         "lights": "CitiStat_Dirtyst_Clogged.ipynb",
         "fin": "City_Liquor_Board_Permits.ipynb",
         "homtax": "CityFinance_histaxhometaxowntax.ipynb",
         "liquor": "City_Liquor_Board_Permits.ipynb",
         "tanf": "DHR_SNAP_TANF_Create.ipynb",
         "snap": "DHR_SNAP_TANF_Create.ipynb",
         "fdf": "TreeBaltimore_TreePlnt.ipynb",
         "findf": "TreeBaltimore_TreePlnt.ipynb",
         "faresCsa": "FARES_Salpr_Shome.ipynb",
         "salepr": "FARES_Salpr_Shome.ipynb",
         "shomes": "FARES_Salpr_Shome.ipynb",
         "banks": "FDIC_Banks.ipynb",
         "df": "HUD.ipynb",
         "cdf": "HUD.ipynb",
         "cdf2": "HUD.ipynb",
         "totemp": "Infousa.ipynb",
         "sml": "Infousa.ipynb",
         "smlbus": "Infousa.ipynb",
         "biz1": "Infousa.ipynb",
         "biz2": "Infousa.ipynb",
         "biz4": "Infousa.ipynb",
         "neiind": "Infousa.ipynb",
         "neibus": "Infousa.ipynb",
         "neiemp": "Infousa.ipynb",
         "totalres": "Mdprop_Permits_Fore_Vac Vio.ipynb",
         "ownroc": "Mdprop_Permits_Fore_Vac Vio.ipynb",
         "vacantsCsa": "Mdprop_Permits_Fore_Vac Vio.ipynb",
         "baltvac": "Mdprop_Permits_Fore_Vac Vio.ipynb",
         "vio": "Mdprop_Permits_Fore_Vac Vio.ipynb",
         "start_date": "Mdprop_Permits_Fore_Vac Vio.ipynb",
         "end_date": "Mdprop_Permits_Fore_Vac Vio.ipynb",
         "mask": "Mdprop_Permits_Fore_Vac Vio.ipynb",
         "mask1": "Mdprop_Permits_Fore_Vac Vio.ipynb",
         "mask2": "Mdprop_Permits_Fore_Vac Vio.ipynb",
         "fore": "Mdprop_Permits_Fore_Vac Vio.ipynb",
         "resrehab": "Mdprop_Permits_Fore_Vac Vio.ipynb",
         "use": "Mdprop_Permits_Fore_Vac Vio.ipynb",
         "demper": "Mdprop_Permits_Fore_Vac Vio.ipynb",
         "filter": "Mdprop_Permits_Fore_Vac Vio.ipynb",
         "constper": "Mdprop_Permits_Fore_Vac Vio.ipynb",
         "comprop": "Mdprop_Permits_Fore_Vac Vio.ipynb",
         "crehab": "Mdprop_Permits_Fore_Vac Vio.ipynb",
         "vsDom": "RBIntel.ipynb",
         "cashsa": "RBIntel.ipynb",
         "reosa": "RBIntel.ipynb"}

modules = ["utils.py",
           "closecrawl.py",
           "tidyaddr.py",
           "indicators.py",
           "create.py",
           "acsDownload.py",
           "BCPSS.py",
           "bpd.py",
           "bidbaltimore.py",
           "citistat.py",
           "cityfinance.py",
           "liquor.py",
           "dhr.py",
           "enoch.py",
           "fares.py",
           "fdic.py",
           "HUD.py",
           "infousa.py",
           "mdprop.py",
           "rbintel.py",
           "treebaltimore.py"]

doc_url = "https://bnia.github.io/VitalSigns/"

git_url = "https://github.com/bnia/VitalSigns/tree/main/"

def custom_doc_links(name): return None



# AUTOGENERATED! DO NOT EDIT! File to edit: notebooks/Infousa.ipynb (unless otherwise specified).

__all__ = ['totemp', 'totemp', 'totemp', 'totemp', 'sml', 'smlbus', 'smlbus', 'smlbus', 'smlbus', 'smlbus', 'smlbus',
           'smlbus', 'smlbus', 'biz1', 'biz1', 'biz1', 'biz1', 'biz1', 'biz2', 'biz2', 'biz2', 'biz2', 'biz2', 'biz4',
           'biz4', 'biz4', 'biz4', 'biz4', 'neiind', 'neiind', 'neiind', 'neiind', 'neiind', 'neiind', 'neibus',
           'neibus', 'neibus', 'neibus', 'neibus', 'neibus', 'neiemp', 'neiemp', 'neiemp', 'neiemp', 'neiemp', 'neiemp']

# Cell
infoUsaCsaTotals.to_csv('numbus18.csv', index=False)

# Cell
# Aggregate Numeric Values by Sum
totemp = infoUsaCsa.groupby('CSA2010')[ ['CSA2010','empl_size'] ].sum(numeric_only=True)
totemp = totemp.merge( csa[ ['CSA2010','tpop10'] ], left_on='CSA2010', right_on='CSA2010' )
totemp = totemp.append( {'CSA2010': 'Baltimore City' , 'tpop10' : 620961, 'empl_size': totemp['empl_size'].sum() }, ignore_index=True)
totemp['totemp'] = totemp['empl_size']
totemp = totemp.drop('empl_size', axis=1)
totemp.tail()

# Cell
sml = infoUsaCsa.copy()
smlbus = sml[ ( sml['empl_rng'].isin(['1 to 4']) ) ]
smlbus.to_csv('smlbus_empl_rng1 to 4.csv')
print('empl_rng 1 to 4: ', smlbus.size / len(smlbus.columns) )

# Cell
smlbus = sml[ ( sml['empl_rng'].isin(['5 to 9']) ) ]
smlbus.to_csv('smlbus_empl_rng5 to 9.csv')
print('empl_rng 5 to 9: ', smlbus.size / len(smlbus.columns) )

# Cell
smlbus = sml[ ( sml['empl_rng'].isin(['10 to 19']) ) ]
smlbus.to_csv('smlbus_empl_rng10 to 19.csv')
print('empl_rng 10 to 19: ', smlbus.size / len(smlbus.columns) )

# Cell
smlbus = sml[ ( sml['empl_rng'].isin(['20 to 49']) ) ]
smlbus.to_csv('smlbus_empl_rng20 to 49.csv')
print('empl_rng 20 to 49: ', smlbus.size / len(smlbus.columns) )

# Cell
# Filter for small businesses
smlbus = sml[ ( sml['empl_rng'].isin(['1 to 4', '5 to 9',  '10 to 19', '20 to 49']) ) ]
smlbus.to_csv('smlbus18_filtered_points.csv')
print('empl_rng 1 to 49: ', smlbus.size / len(smlbus.columns) )

# Cell
# Aggregate Numeric Values by Sum
smlbus['smlbus'] = 1
smlbus = smlbus.groupby('CSA2010')[ ['CSA2010','smlbus'] ].sum(numeric_only=True)
smlbus = smlbus.merge( csa[ ['CSA2010','tpop10'] ], left_on='CSA2010', right_on='CSA2010' )
smlbus = smlbus.append( {'CSA2010': 'Baltimore City' , 'tpop10' : 620961, 'smlbus': gdf2['smlbus'].sum() }, ignore_index=True)
smlbus.tail()

# Cell
# 145 -biz1XX

# Filter for small businesses
biz1 = infoUsaCsa[ ( infoUsaCsa['first_year'].isin( ['2018'] ) ) ]
print('Count: first_year == 2018: ', biz1.size / len(biz1.columns) )
biz1 = biz1[ ['CSA2010'] ]
#numerator.to_csv('biz18_numerator_csasWithCounts.csv')
biz1['biz1Count'] = 1

# Cell
# Aggregate Numeric Values by Sum
biz1 = biz1.groupby('CSA2010').sum(numeric_only=True)
biz1 = biz1.merge( csa[ ['CSA2010','tpop10'] ], left_on='CSA2010', right_on='CSA2010' )
biz1 = biz1.append( {'CSA2010': 'Baltimore City' , 'tpop10' : 620961, 'biz1Count': biz1['biz1Count'].mean() }, ignore_index=True)
biz1.tail(1)

# Cell
# Create the Indicator
biz1['biz1'] = biz1['biz1Count'] / infoUsaCsaTotals['numbus']
biz1.head()

# Cell
# 151 - biz2XX

# Filter for small businesses
biz2 = infoUsaCsa[ ( infoUsaCsa['first_year'].isin( ['2016', '2017', '2018'] ) ) ]
print('Count: first_year == 2018, 2017, 2016: ', biz2.size / len(biz2.columns) )
biz2 = biz2[ ['CSA2010'] ]
#numerator.to_csv('biz18_numerator_csasWithCounts.csv')
biz2['biz2Count'] = 1

# Cell
# Aggregate Numeric Values by Sum
biz2 = biz2.groupby('CSA2010').sum(numeric_only=True)
biz2 = biz2.merge( csa[ ['CSA2010','tpop10'] ], left_on='CSA2010', right_on='CSA2010' )
biz2 = biz2.append( {'CSA2010': 'Baltimore City' , 'tpop10' : 620961, 'biz2Count': biz2['biz2Count'].mean() }, ignore_index=True)
biz2.tail(1)

# Cell
# 152 - biz4XX

# Filter for small businesses
biz4 = infoUsaCsa[ ( infoUsaCsa['first_year'].isin( ['2015', '2016', '2017', '2018'] ) ) ]
print('Count: first_year == 2018, 2017, 2016, 2015: ', biz2.size / len(biz2.columns) )
biz4 = biz4[ ['CSA2010'] ]
#numerator.to_csv('biz18_numerator_csasWithCounts.csv')
biz4['biz4Count'] = 1

# Cell
# Aggregate Numeric Values by Sum
biz4 = biz4.groupby('CSA2010').sum(numeric_only=True)
biz4 = biz4.merge( csa[ ['CSA2010','tpop10'] ], left_on='CSA2010', right_on='CSA2010' )
biz4 = biz4.append( {'CSA2010': 'Baltimore City' , 'tpop10' : 620961, 'biz4Count': biz4['biz4Count'].mean() }, ignore_index=True)
biz4.tail(1)

# Cell
# 157 - neiindXX

# Filter for small businesses
neiind = infoUsaCsa.copy()
neiind['naics_extra_short'] = neiind.prim_naics.astype(str).str[:-6].astype(np.int64)
neiind = infoUsaCsa[ ( neiind['naics_extra_short'].isin( [44, 45, 52, 54, 62, 71, 72, 81] ) ) ]
print('Count of Naics Starting With: 44, 45, 52, 54, 62, 71, 72, 81: ', neiind.size / len(neiind.columns) )
neiind = neiind[ ['CSA2010'] ]
#numerator.to_csv('biz18_numerator_csasWithCounts.csv')
neiind['neiind'] = 1

# Cell
# Aggregate Numeric Values by Sum
neiind = neiind.groupby('CSA2010').sum(numeric_only=True)
neiind = neiind.merge( csa[ ['CSA2010','tpop10'] ], left_on='CSA2010', right_on='CSA2010' )
neiind = neiind.append( {'CSA2010': 'Baltimore City' , 'tpop10' : 620961, 'neiind': neiind['neiind'].sum() }, ignore_index=True)
neiind.tail(1)

# Cell
# 158 - neibus

# Filter for small businesses
neibus = infoUsaCsa.copy()
neibus['naics_extra_short'] = neibus.prim_naics.astype(str).str[:-6].astype(np.int64)
neibus = infoUsaCsa[ ( neibus['naics_extra_short'].isin( [44, 45, 52, 54, 62, 71, 72, 81] ) ) ]
print('Count of Naics Starting With: 44, 45, 52, 54, 62, 71, 72, 81: ', neibus.size / len(neibus.columns) )
neibus = neibus[ ['CSA2010'] ]
#numerator.to_csv('biz18_numerator_csasWithCounts.csv')
neibus['neibus'] = 1
neibus.head()

# Cell
# Aggregate Numeric Values by Sum
neibus = neibus.groupby('CSA2010').sum(numeric_only=True)
neibus = neibus.merge( csa[ ['CSA2010','tpop10'] ], left_on='CSA2010', right_on='CSA2010' )
neibus = neibus.append( {'CSA2010': 'Baltimore City' , 'tpop10' : 620961, 'neibus': neibus['neibus'].sum() }, ignore_index=True)
neibus['neibus'] = neibus['neibus'] * 1000 / neibus['tpop10']
neibus.tail(1)

# Cell
# 159 - neiempXX

# Filter for small businesses
neiemp = infoUsaCsa.copy()
neiemp['naics_extra_short'] = neiemp.prim_naics.astype(str).str[:-6].astype(np.int64)
neiemp = infoUsaCsa[ ( neiemp['naics_extra_short'].isin( [44, 45, 52, 54, 62, 71, 72, 81] ) ) ]
print('Count of Naics Starting With: 44, 45, 52, 54, 62, 71, 72, 81: ', neiemp.size / len(neiemp.columns) )
#numerator.to_csv('biz18_numerator_csasWithCounts.csv')

# Cell
# Aggregate Numeric Values by Sum
neiemp = neiemp.groupby('CSA2010')[ ['CSA2010','empl_size'] ].sum(numeric_only=True)
neiemp = neiemp.merge( csa[ ['CSA2010','tpop10'] ], left_on='CSA2010', right_on='CSA2010' )
neiemp = neiemp.append( {'CSA2010': 'Baltimore City' , 'tpop10' : 620961, 'empl_size': neiemp['empl_size'].sum() }, ignore_index=True)
neiemp['neiemp'] = neiemp['empl_size']
neiemp = neiemp.drop('empl_size', axis=1)
neiemp.tail()


# AUTOGENERATED! DO NOT EDIT! File to edit: notebooks/FARES_Salpr_Shome.ipynb (unless otherwise specified).

__all__ = ['faresCsa', 'salepr', 'salepr', 'salepr', 'salepr', 'salepr', 'faresCsa', 'shomes', 'shomes', 'shomes',
           'shomes']

# Cell
# Copy the Data
faresCsa = fares.copy()
faresCsa['SALEDATE'] = pd.to_datetime(faresCsa['SALEDATE'], format='%Y%m%d') # ,infer_datetime_format=True)

# Query the Data
salepr = faresCsa[
  (faresCsa['PROPERTYIN'].str.contains('CONDOMINIUM|SINGLE FAMILY', regex=True) )
& (faresCsa['SALEDATE'] >= '20'+year+'-01-01')
& (faresCsa['SALEDATE'] <= '20'+year+'-12-31')
& (faresCsa['SALEAMOUNT'] > 1000 )
& (faresCsa['PRIMARYCAT'] == 'ARMS LENGTH' )
]

# Prep and Save the Filtered Records
salepr.rename(columns={ 'SALEAMOUNT':'29-salepr'+year}, inplace=True)
salepr.drop(columns=['geometry', 'PRIMARYCAT', 'PROPERTYIN', 'LANDUSECOD', 'SALEDATE' ], inplace=True)
salepr.to_csv('fares_filtered_'+year+'.csv')

# *Special*: # UPDATE HERE AND THEN GROUP
salepr = salepr.append({'CSA2010': 'Baltimore City' , '29-salepr'+year : salepr['29-salepr'+year].median() } , ignore_index=True)
salepr = salepr.groupby('CSA2010').median(numeric_only=True)

# Make sure ALL csas and BaltimoreCity are included and sorted.
salepr = csa.merge( salepr, left_on='CSA2010', right_on='CSA2010', how='outer' )
salepr.drop(columns=['OBJECTID', 'Shape__Length', 'Shape__Area', 'geometry' ], inplace=True)
salepr = salepr[['CSA2010', '29-salepr'+year]]

display( salepr.head(2) )
salepr.tail(2)

salepr.to_csv('29-salepr'+year+'.csv')

# & (faresCsa['PRIMCATCOD'] == 'ARMS LENGTH' )
# faresCsa['LANDUSECOD'].str.contains('HIGH RISE CONDO|CONDOMINIUM|RESIDENTIAL (NEC)|MID RISE CONDO|MULTI FAMILY DWELLING|MULTI FAMILY 10 UNITS LESS|SFR|TOWNHOUSE/ROWHOUSE', regex=False)

# Cell
# Copy the Data
faresCsa = fares.copy()
faresCsa['SALEDATE'] = pd.to_datetime(faresCsa['SALEDATE'], format='%Y%m%d') # ,infer_datetime_format=True)

# Query the Data
shomes = faresCsa[
  (faresCsa['PROPERTYIN'].str.contains('CONDOMINIUM|SINGLE FAMILY', regex=True) )
& (faresCsa['SALEDATE'] >= '20'+year+'-01-01')
& (faresCsa['SALEDATE'] <= '20'+year+'-12-31')
& (faresCsa['SALEAMOUNT'] > 1000 )
& (faresCsa['PRIMARYCAT'] == 'ARMS LENGTH' )
]

# Prep and Save the Filtered Records
shomes['31-shomes'+year] = 1
# shomes.rename(columns={ 'SALEAMOUNT':'29-shomes'+year}, inplace=True)
# shomes.drop(columns=['geometry', 'PRIMARYCAT', 'PROPERTYIN', 'LANDUSECOD', 'SALEDATE', 'SALEAMOUNT'], inplace=True)
shomes = shomes[['CSA2010','31-shomes'+year]]
shomes.to_csv('fares_filtered_'+year+'.csv')

# *Special*: # UPDATE HERE AND THEN GROUP
# shomes = shomes.append({'CSA2010': 'Baltimore City', '31-shomes'+year : shomes['31-shomes'+year].sum() } , ignore_index=True)
shomes = shomes.groupby('CSA2010').sum(numeric_only=True)

# Make sure ALL csas and BaltimoreCity are included and sorted.
shomes = csa.merge( shomes, left_on='CSA2010', right_on='CSA2010', how='outer' )
shomes.drop(columns=['OBJECTID', 'Shape__Length', 'Shape__Area', 'geometry' ], inplace=True)

# *SPECIAL* Update the baltimore CSA.
shomes.at[55] = shomes.sum(numeric_only=True)
shomes.at[55, 'CSA2010'] = 'Baltimore City'

display( shomes.head(2) )
shomes.tail(2)

shomes.to_csv('31-shomes'+year+'.csv')

# & (faresCsa['PRIMCATCOD'] == 'ARMS LENGTH' )
# faresCsa['LANDUSECOD'].str.contains('HIGH RISE CONDO|CONDOMINIUM|RESIDENTIAL (NEC)|MID RISE CONDO|MULTI FAMILY DWELLING|MULTI FAMILY 10 UNITS LESS|SFR|TOWNHOUSE/ROWHOUSE', regex=False)


# AUTOGENERATED! DO NOT EDIT! File to edit: notebooks/Bidbalt_TaxSale_Taxlien.ipynb (unless otherwise specified).

__all__ = ['taxlien']

# Cell
def taxlien(df, totalres, year):
  taxlien = df.copy()
  # Aggregate Numeric Values by Sum
  taxlien['taxlienCount'] = 1
  taxlien = taxlien.groupby('CSA2010').sum(numeric_only=True)

  # Make sure ALL csas and BaltimoreCity are included. among other things
  taxlien = totalres[ ['CSA2010','totalres'+year] ].merge( taxlien, left_on='CSA2010', right_on='CSA2010', how='outer' )

  # Update the baltimore CSA.
  taxlien.at[55,'taxlienCount'] = taxlien['taxlienCount'].sum()

  # Create the Indicator
  taxlien['40-taxlien'+year] = taxlien['taxlienCount'] / totalres['totalres'+year] * 100

  taxlien = taxlien[['CSA2010', 'taxlienCount', '40-taxlien'+year, 'totalres'+year ]]

  compareYears = gpd.read_file("https://services1.arcgis.com/mVFRs7NF4iFitgbY/ArcGIS/rest/services/Taxlien/FeatureServer/0/query?where=1%3D1&outFields=*&returnGeometry=true&f=pgeojson");
  prevYear = 'taxlien'+ str( int(year) - 1 )
  if prevYear in compareYears.columns:
    taxlien = taxlien.merge( compareYears[['CSA2010', prevYear]], left_on='CSA2010', right_on='CSA2010', how='outer' )
    taxlien['change'] = taxlien['40-taxlien'+year] - taxlien[ prevYear ]
    taxlien['percentChange'] = taxlien['change'] / taxlien[ prevYear ] * 100
    taxlien['change'] = taxlien['change'].apply(lambda x: "{:.2f}".format(x) )
    taxlien = taxlien[['CSA2010', 'taxlienCount', 'totalres19', 'taxlien18', '40-taxlien19', 'percentChange', 'change']]

  return taxlien


# AUTOGENERATED! DO NOT EDIT! File to edit: notebooks/EnochPratt_Libcard.ipynb (unless otherwise specified).

__all__ = ['fdf', 'findf', 'findf', 'findf']

# Cell
%%capture
# export
# @title Run: Import Modules

# These imports will handle everything
import os
import sys
import csv
from IPython.display import clear_output
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import geopandas as gpd
from geopandas import GeoDataFrame
import psycopg2
import pyproj
from pyproj import Proj, transform
# conda install -c conda-forge proj4
from shapely.geometry import Point
from shapely.geometry import LineString
from shapely import wkb
from shapely.wkt import loads
# https://pypi.org/project/geopy/
from geopy.geocoders import Nominatim
import folium
from folium import plugins

from dataplay.merge import mergeDatasets

import dexplot as dxp

# In case file is KML, enable support
import fiona
fiona.drvsupport.supported_drivers['kml'] = 'rw'
fiona.drvsupport.supported_drivers['KML'] = 'rw'

# Cell
fdf = geoms.workWithGeometryData(method='pinp', df=gdf, polys=csa_gdf, ptsCoordCol='geometry', polygonsCoordCol='geometry', polyColorCol=False, polygonsLabel='RecordsInCsa', pntsClr='red', polysClr='white')
findf = fdf.copy()
findf = findf[['CSA2010', 'tpop10', 'pointsinpolygon', 'geometry']]
findf['libcard19'] = findf['pointsinpolygon'] / findf['tpop10'] * 1000

# Cell
findf = findf.append({
    'CSA2010': 'Baltimore City',
    'tpop10': findf['tpop10'].sum(),
    'pointsinpolygon': findf['pointsinpolygon'].sum(),
    'geometry': 'na',
    'libcard19': (findf['pointsinpolygon'].sum() / findf['tpop10'].sum() * 1000), }, ignore_index=True)

# Cell
%%capture
# export
# @title Run: Import Modules

# These imports will handle everything
import os
import sys
import csv
from IPython.display import clear_output
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import geopandas as gpd
from geopandas import GeoDataFrame
import psycopg2
import pyproj
from pyproj import Proj, transform
# conda install -c conda-forge proj4
from shapely.geometry import Point
from shapely.geometry import LineString
from shapely import wkb
from shapely.wkt import loads
# https://pypi.org/project/geopy/
from geopy.geocoders import Nominatim
import folium
from folium import plugins

from dataplay.merge import mergeDatasets

import dexplot as dxp

# In case file is KML, enable support
import fiona
fiona.drvsupport.supported_drivers['kml'] = 'rw'
fiona.drvsupport.supported_drivers['KML'] = 'rw'


# AUTOGENERATED! DO NOT EDIT! File to edit: notebooks/RBIntel.ipynb (unless otherwise specified).

__all__ = ['vsDom', 'cashsa', 'reosa']

# Cell
import numpy as np
import pandas as pd
# import geopandas
import geopandas as gpd
# from geopandas import GeoDataFrame

# Cell
def vsDom(df, yr):
  id = '30'
  shortname = 'dom'
  fincol = id+'-'+shortname+year

  # Create the Numerator and Denominator
  numer = df[['DaysonMark','CSA2010']].copy()

  # Filter Em
  numer = numer[ numer['DaysonMark'] > 0]
  print( numer.shape[0] )

  # Get Bcity Val
  bCityVal = numer.median(numeric_only=True)['DaysonMark']

  # Group by CSA
  numer = numer.groupby('CSA2010').median(numeric_only=True) # use .median to calculate DOM.

  # Make sure ALL csas and BaltimoreCity are included and sorted.
  numer = csa.merge( numer, left_on='CSA2010', right_on='CSA2010', how='outer' )
  numer.drop( columns=['geometry', 'Shape__Length','Shape__Area', 'OBJECTID', 'tpop10'], inplace=True)

  # Bcity is the median of all the records an not the community medians.
  # Incorrect Bcity median IFF Groupby keeps a 'False' row (index 56)
  numer.at[55,'DaysonMark']= bCityVal

  # Perform the calculation
  numer[fincol] = numer['DaysonMark']

  compareYears = gpd.read_file("https://services1.arcgis.com/mVFRs7NF4iFitgbY/ArcGIS/rest/services/"+shortname.capitalize()+"/FeatureServer/0/query?where=1%3D1&outFields=*&returnGeometry=true&f=pgeojson");
  goback = 2 if year == '19' else 3 if year == '20' else 1
  prevYear = shortname + str( int(year) - goback )
  if prevYear in compareYears.columns:
    numer = numer.merge( compareYears[['CSA2010', prevYear]], left_on='CSA2010', right_on='CSA2010', how='outer' )
    numer['change'] = numer[id+'-'+shortname+year] - numer[ prevYear ]
    numer['percentChange'] = numer['change' ] / numer[ prevYear ] * 100
    numer['change'] = numer['change'].apply(lambda x: "{:.2f}".format(x) )
    print( 'Records Matching Query: ', numer.size / len(numer.columns) )

  return numer

# Cell
def cashsa(df, yr):
  id = '38'
  shortname = 'cashsa'
  fincol = id+'-'+shortname+year

  # Create the Numerator and Denominator
  numer = df[['BuyerFinan','CSA2010']].copy()
  numer['count'] = 1
  denom = numer.copy()

  # Filter Em
  numer = numer[ numer['BuyerFinan'].str.contains('.Cash.|.Cash|Cash.|Cash', regex=True, na=False) ]
  print("LENGTH AFTER FILTER: ", len(numer) )

  # Get Bcity Val
  bCityVal = numer.sum(numeric_only=True)['count']
  bCityValDenom = denom.sum(numeric_only=True)['count']

  # Group by CSA
  numer = numer.groupby('CSA2010').sum(numeric_only=True)
  denom = denom.groupby('CSA2010').sum(numeric_only=True)

  # Make sure ALL csas and BaltimoreCity are included and sorted.
  numer = csa.merge( numer, left_on='CSA2010', right_on='CSA2010', how='outer' )
  numer.drop( columns=['geometry', 'Shape__Length','Shape__Area', 'OBJECTID', 'tpop10'], inplace=True)

  denom = csa.merge( denom, left_on='CSA2010', right_on='CSA2010', how='outer' )
  denom.drop( columns=['geometry', 'Shape__Length','Shape__Area', 'OBJECTID', 'tpop10'], inplace=True)

  # Bcity is the sum of the community sums.
  # Incorrect Bcity Sum IFF Groupby keeps a 'False' row (index 56)
  numer.at[55,'count']= bCityVal
  denom.at[55,'count']= bCityValDenom

  # Perform the calculation
  numer[fincol] = numer['count'] / denom['count'] * 100

  compareYears = gpd.read_file("https://services1.arcgis.com/mVFRs7NF4iFitgbY/ArcGIS/rest/services/"+shortname.capitalize()+"/FeatureServer/0/query?where=1%3D1&outFields=*&returnGeometry=true&f=pgeojson");
  goback = 2 if year == '19' else 3 if year == '20' else 1
  prevYear = shortname + str( int(year) - goback )
  if prevYear in compareYears.columns:
    numer = numer.merge( compareYears[['CSA2010', prevYear]], left_on='CSA2010', right_on='CSA2010', how='outer' )
    numer['change'] = numer[id+'-'+shortname+year] - numer[ prevYear ]
    numer['percentChange'] = numer['change' ] / numer[ prevYear ] * 100
    numer['change'] = numer['change'].apply(lambda x: "{:.2f}".format(x) )
    print( 'Records Matching Query: ', numer.size / len(numer.columns) )

  return numer

# Cell
def reosa(df, yr):
  id = '39'
  shortname = 'reosa'
  fincol = id+'-'+shortname+year

  # Create the Numerator and Denominator
  numer = df[['Foreclosur','CSA2010']].copy()
  numer['count'] = 1

  denom = numer.copy()
  denom['count'] = 1

  # Filter Em
  numer = numer[ numer['Foreclosur'].str.contains('.Y.|.Y|Y.|Y', regex=True, na=False) ]
  print( numer['Foreclosur'].value_counts() )

  # Get Bcity Val
  bCityVal = numer.sum(numeric_only=True)['count']
  bCityValDenom = denom.sum(numeric_only=True)['count']

  # Group by CSA
  numer = numer.groupby('CSA2010').sum(numeric_only=True)
  denom = denom.groupby('CSA2010').sum(numeric_only=True)

  # Make sure ALL csas and BaltimoreCity are included and sorted.
  numer = csa.merge( numer, left_on='CSA2010', right_on='CSA2010', how='outer' )
  numer.drop( columns=['geometry', 'Shape__Length','Shape__Area', 'OBJECTID', 'tpop10'], inplace=True)

  denom = csa.merge( denom, left_on='CSA2010', right_on='CSA2010', how='outer' )
  denom.drop( columns=['geometry', 'Shape__Length','Shape__Area', 'OBJECTID', 'tpop10'], inplace=True)

  # Bcity is the sum of the community sums.
  # Incorrect Bcity Sum IFF Groupby keeps a 'False' row (index 56)
  numer.at[55,'count']= bCityVal
  denom.at[55,'count']= bCityValDenom

  # Perform the calculation
  numer['denomCount'] = denom['count']
  numer[fincol] = numer['count'] / numer['denomCount'] * 100

  compareYears = gpd.read_file("https://services1.arcgis.com/mVFRs7NF4iFitgbY/ArcGIS/rest/services/"+shortname.capitalize()+"/FeatureServer/0/query?where=1%3D1&outFields=*&returnGeometry=true&f=pgeojson");
  goback = 2 if year == '19' else 3 if year == '20' else 1
  prevYear = shortname + str( int(year) - goback )
  if prevYear in compareYears.columns:
    numer = numer.merge( compareYears[['CSA2010', prevYear]], left_on='CSA2010', right_on='CSA2010', how='outer' )
    numer['change'] = numer[id+'-'+shortname+year] - numer[ prevYear ]
    numer['percentChange'] = numer['change' ] / numer[ prevYear ] * 100
    numer['change'] = numer['change'].apply(lambda x: "{:.2f}".format(x) )
    print( 'Records Matching Query: ', numer.size / len(numer.columns) )

  return numer


# AUTOGENERATED! DO NOT EDIT! File to edit: notebooks/01_Close Crawl.ipynb (unless otherwise specified).


from __future__ import absolute_import, print_function, unicode_literals



from __future__ import absolute_import, print_function, unicode_literals



from __future__ import absolute_import, print_function, unicode_literals



from __future__ import absolute_import, print_function, unicode_literals



from __future__ import absolute_import, print_function, unicode_literals



from __future__ import absolute_import, print_function, unicode_literals


__all__ = ['HEADER', 'URL', 'CASE_PAT', 'CASE_TYPES', 'HTML_DIR', 'HTML_FILE', 'CHECKPOINT', 'NO_CASE', 'FEATURES',
           'FIELDS', 'INTERNAL_FIELDS', 'Session', 'Spider', 'scrape', 'filter_addr', 'PUNCTUATION', 'street_address',
           'TITLE_SPLIT_PAT', 'ZIP_STR', 'ZIP_PAT', 'MONEY_STR', 'MONEY_PAT', 'NULL_ADDR', 'STRIP_ADDR', 'Miner',
           'FEATURES', 'FIELDS', 'INTERNAL_FIELDS', 'temp_output', 'file_array', 'top5', 'Cleaner', 'temp_output',
           'output', 'df_obj', 'close_crawl', 'args']

# Cell
import mechanicalsoup
mechanicalsoup.__version__

# Cell
"""settings

Configuration settings and global variables for the entire project. This file
is intended to only be used as a non-executable script.
"""

from os import path

# browser settings
HEADER = ("Mozilla/5.0 (X11; U; Linux i686; en-US; rv:1.9.0.1)"
          " Gecko/2008071615 Fedora/3.0.1-1.fc9 Firefox/3.0.1")
URL = "http://casesearch.courts.state.md.us/casesearch/"
CASE_PAT = "24{type}{year}00{num}"

# scraping parameters
CASE_TYPES = ('O', 'C')

# temporary directory settings
HTML_DIR = "responses"
HTML_FILE = path.join(HTML_DIR, "{case}")

# log file settings
CHECKPOINT = "checkpoint.json"
NO_CASE = "no_case.json"

# data mining settings
FEATURES = [
    "Filing Date",
    "Case Number",
    "Case Type",
    "Title",
    "Plaintiff",
    "Defendant",
    "Address",
    "Business or Organization Name",
    "Party Type",
]
FIELDS = FEATURES + [ "Zip Code", "Partial Cost" ]
INTERNAL_FIELDS = [ "Business or Organization Name", "Party Type"]

# Cell
"""local_browser

NOTICE: Close Crawl runs its browser form submissions through Mechanize.
The module, however, is deprecated and does not support Python 3. The more
stable and maintained Mechanize and BeautifulSoup wrapper, MechanicalSoup,
will be replacing the Mechanize methods to support Python 3.

This module contains the configurations and settings for the browser used for
crawling and scraping through the pages in Close Crawl. The script contains the
implementation of the Session class which inherits attributes from the classobj
mechanize.Browser()

The script works as an internal module for Close Crawl, but can be imported
as a module for testing purposes.

TODO:
    Replace deprecated Mechanize with MechanicalSoup
    Fork Mechanize to support Python3

"""

# Cell
#nbdev_comment from __future__ import absolute_import, print_function, unicode_literals
# import cookielib
import http.cookiejar as cookielib # for Python3
import warnings
from urllib.request import urlopen
# from urllib import urlopen urllib.request

## from mechanize import Browser, _http
import mechanicalsoup

# from settings import HEADER, URL

warnings.filterwarnings("ignore", category=UserWarning)

# Cell
class Session(object):

    def __init__(self):
        """Constructor

        Args:
            None

        Attributes:
            browser (`mechanize._mechanize.Browser`): browser object in session
        """

        self.browser = mechanicalsoup.StatefulBrowser()

        # set error and debug handlers for the browser

        # cookie jar
        self.browser.set_cookiejar(cookielib.LWPCookieJar())

        # browser options
        # self.browser.set_handle_equiv(True)
        # self.browser.set_handle_gzip(True)
        # self.browser.set_handle_redirect(True)
        # self.browser.set_handle_referer(True)
        # self.browser.set_handle_robots(False)

        # follows refresh 0 but doesn't hang on refresh > 0
        #self.browser.set_handle_refresh( _http.HTTPRefreshProcessor(), max_time=1 )

        # user-Agent
        # self.browser.addheaders = [("User-agent", HEADER)]

    def close(self):
        """Destructor for Session. Closes current browser session

        Args:
            None

        Returns:
            None
        """
        self.browser.close()

    def case_id_form(self, case):
        """Grabs the form in the case searching page, and inputs the
        case number to return the response.

        Args:
            case (`str`): case ID to be scraped

        Returns:
            response (`str`): HTML response
        """

        # iterate through the forms to find the correct one
        #for form in self.browser.forms():
        #    if form.attrs["name"] == "inquiryFormByCaseNum":
        #        self.browser.form = form
        #        break

        self.browser.select_form('form[action="/casesearch/inquiryByCaseNum.jis"]')

        # submit case ID and return the response
        self.browser["caseId"] = case
        response = self.browser.submit_selected()
        response = response.text
        # if any( case_type in response.upper() for case_type in ("FORECLOSURE", "FORECLOSURE RIGHTS OF REDEMPTION", "MOTOR TORT") ): print (response.upper)

        self.browser.open("http://casesearch.courts.state.md.us/casesearch/inquiryByCaseNum.jis")
        # , "MOTOR TORT"
        return response if any(
            case_type in response.upper() for case_type in
            ("FORECLOSURE", "FORECLOSURE RIGHTS OF REDEMPTION")
        ) else False

    def disclaimer_form(self):
        """Navigates to the URL to proceed to the case searching page

        Args:
            None

        Returns:
            None
        """

        # visit the site
        print(URL)
        self.browser.open("http://casesearch.courts.state.md.us/casesearch/")

        # select the only form on the page
        self.browser.select_form('form')

        # select the checkbox
        self.browser["disclaimer"] = ['Y']

        # submit the form
        self.browser.submit_selected()

    @staticmethod
    def server_running():
        """Checks the status of the Casesearch servers

        Args:
            None

        Returns:
            `True` if server is up, `False` otherwise
        """
        return urlopen(URL).getcode() == 200


# Cell
#nbdev_comment from __future__ import absolute_import, print_function, unicode_literals
from json import dumps, load
from os import path, makedirs
from random import uniform
import sys
from time import sleep

from tqdm import trange

# from local_browser import Session
# from settings import CASE_PAT, CHECKPOINT, HTML_DIR, HTML_FILE

# Cell
class Spider(object):

    def __init__(self, case_type, year, bounds=range(1, 6), gui=False):

        # initial disclaimer page for terms and agreements
        self.browser = Session()

        if not self.browser.server_running():
            sys.exit("Server is unavailable at the moment")

        print (self.browser)

        self.browser.disclaimer_form()

        self.WAITING_TIME = 0
        self.case_type = case_type
        self.year = year
        self.bounds = bounds

        if not path.exists(HTML_DIR):
            makedirs(HTML_DIR)

    def save_response(self):

        case_range = trange(
            len(self.bounds), desc="Crawling", leave=True
        )

        for case_num in case_range:

            if case_num and not case_num % 500:
                print("500 CASES SCRAPED. SCRIPT WILL WAIT 5 MINUTES TO RESUME")

                for i in range(150, 0, -1):
                    sleep(1)
                    sys.stdout.write('\r' + "%02d:%02d" % divmod(i, 60))
                    sys.stdout.flush()

            case = CASE_PAT.format(
                type=self.case_type,
                year=self.year,
                num="{:04d}".format(int(str(self.bounds[case_num])[-4:]))
            )

            try:

                wait = uniform(0.0, 0.5)
                sleep(wait)

                self.WAITING_TIME += wait

                case_range.set_description("Crawling {}".format(case))

                stripped_html = self.browser.case_id_form(case)
                # print('returend this ' , stripped_html)

                if stripped_html:
                    with open(
                        HTML_FILE.format(case=case) + ".html", 'w'
                    ) as case_file:
                        case_file.write(str(stripped_html))

            # pause process
            except KeyboardInterrupt:

                self.dump_json({
                    "error_case":
                        "{:04d}".format(int(str(self.bounds[case_num])[-4:])),
                        "year": self.year,
                        "type": self.type
                })
                print("Crawling paused at", case)
                break

            # case does not exist
            except IndexError:

                self.dump_json({"error_case": case})
                print(case, "does not exist")
                break

        # close browser and end session
        self.close_sesh()

    @staticmethod
    def dump_json(data):

        with open(CHECKPOINT, "r+") as checkpoint:
            checkpoint_data = load(checkpoint)

            for key, val in data.items():
                checkpoint_data[key] = val

            checkpoint_data[key] = data
            checkpoint.seek(0)
            checkpoint.write(dumps(checkpoint_data))
            checkpoint.truncate()

    def close_sesh(self):
        self.browser.close()


# Cell
scrape = True
if scrape:
  spider = Spider(
    case_type='O', year=20,
    bounds=range(0000, 9000), gui=False
  )
  spider.save_response()
  wait = spider.WAITING_TIME

# Cell
"""Patterns

Regular expression patterns and string filtering functions implemented in the
project. This file is intended to only be used as a non-executable script.

TODO:
    Finish docs

(\d{1,4}\s[\w\s]{1,20}((?:st(reet)?|ln|lane|ave(nue)?|r(?:oa)?d|highway|hwy|
dr(?:ive)?|sq(uare)?|tr(?:ai)l|c(?:our)?t|parkway|pkwy|cir(cle)?|ter(?:race)?|
boulevard|blvd|pl(?:ace)?)\W?(?=\s|$))(\s(apt|block|unit)\W?([A-Z]|\d+))?)
"""

# Cell
from re import compile as re_compile
from re import I as IGNORECASE
from string import punctuation

# Cell
PUNCTUATION = punctuation.replace('#', '')  # all punctuations except '#'

street_address = re_compile(
    "("  # begin regex group
    "\d{1,4}\s"  # house number
    "[\w\s]{1,20}"  # street name
    "("  # start street type group
    "(?:st(reet)?|ln|lane|ave(nue)?"  # (st)reet, lane, ln, (ave)nue
    "|r(?:oa)?d|highway|hwy|dr(?:ive)?"  # rd, road, hwy, highway, (dr)ive
    "|sq(uare)?|tr(?:ai)l|c(?:our)?t"  # (sq)uare, (tr)ail, ct, court
    "|parkway|pkwy|cir(cle)?|ter(?:race)?"  # parkway, pkwy, (cir)cle, (ter)race
    "|boulevard|blvd|pl(?:ace)?"  # boulevard, bvld, (pl)ace
    "\W?(?=\s|$))"  # look ahead for whitespace or end of string
    ")"  # end street type group
    "(\s(apt|block|unit)(\W|#)?([\d|\D|#-|\W])+)?"  # apt, block, unit number
    ")",  # end regex group
    IGNORECASE  # case insensitive flag
)

# case insensitive delimiter for Titles
TITLE_SPLIT_PAT = re_compile(" vs ", IGNORECASE)

# pattern for Baltimore zip codes
ZIP_STR = "2\d{4}"
ZIP_PAT = re_compile(ZIP_STR)

# regex pattern to capture monetary values between $0.00 and $999,999,999.99
# punctuation insensitive
MONEY_STR = "\$\d{,3},?\d{,3},?\d{,3}\.?\d{2}"
MONEY_PAT = re_compile(MONEY_STR)

NULL_ADDR = re_compile(
    "^("
    "(" + MONEY_STR + ")"
    "|(" + ZIP_STR + ")"
    "|(\d+)"
    "|(" + ZIP_STR + ".*" + MONEY_STR + ")"
    ")$",
    IGNORECASE
)

STRIP_ADDR = re_compile(
    "(balto|" + ZIP_STR + "|md|" + MONEY_STR + ").*",
    IGNORECASE
)


def filter_addr(address):

    try:
      return ''.join(
        street_address.search(
          address.translate( str.maketrans('','', PUNCTUATION) )
        ).group(0)
      )
    except AttributeError:
      return ''


# Cell
"""Miner"""
#nbdev_comment from __future__ import absolute_import, print_function, unicode_literals
from csv import DictWriter
from json import dump, dumps, load
from os import path

from bs4 import BeautifulSoup
from tqdm import trange

# from patterns import MONEY_PAT, TITLE_SPLIT_PAT, ZIP_PAT, filter_addr
# from settings import HTML_FILE, NO_CASE

# Cell
# data mining settings
FEATURES = [
    "Filing Date",
    "Case Number",
    "Case Type",
    "Title",
    "Plaintiff",
    "Defendant",
    "Address",
    "Business or Organization Name",
    "Party Type",
]
FIELDS = FEATURES + [ "Zip Code", "Partial Cost" ]
INTERNAL_FIELDS = [ "Business or Organization Name", "Party Type"]

class Miner(object):

    def __init__(self, responses, output, debug=False):

        self.responses = responses
        self.output = output
        self.debug = debug
        self.dataset = []
        self.maybe_tax = False
        self.features = [i + ':' for i in FEATURES]

    def scan_files(self):

        case_range = trange(len(self.responses), desc="Mining", leave=True) \
            if not self.debug else range(len(self.responses))

        for file_name in case_range:
            with open(
                HTML_FILE.format(case=self.responses[file_name]), 'r'
            ) as html_src:

                if not self.debug:
                    case_range.set_description(
                        "Mining {}".format(self.responses[file_name])
                    )

                feature_list = self.scrape(html_src.read())
                row = self.distribute(feature_list)

                if not row:

                    if not path.isfile(NO_CASE):
                        with open(NO_CASE, 'w') as no_case_file:
                            dump([], no_case_file)

                    with open(NO_CASE, "r+") as no_case_file:
                        no_case_data = load(no_case_file)
                        no_case_data.append(str(self.responses[file_name][:-5]))
                        no_case_file.seek(0)
                        no_case_file.write(dumps(sorted(set(no_case_data))))
                        no_case_file.truncate()

                self.dataset.extend(row)

    def export(self):

        file_exists = path.isfile(self.output)

        with open(self.output, 'a') as csv_file:
            writer = DictWriter(
                csv_file,
                fieldnames=[
                    col for col in FIELDS if col not in INTERNAL_FIELDS
                ]
            )

            if not file_exists:
                writer.writeheader()

            for row in self.dataset:
                writer.writerow(row)

    def scrape(self, html_data):
        """Scrapes the desired features

        Args:
            html_data: <str>, source HTML

        Returns:
            scraped_features: <dict>, features scraped and mapped from content
        """

        soup = BeautifulSoup(html_data, "html.parser")

        # Search for the word 'tax in the document'
        if "tax" in soup.text.lower():
            self.maybe_tax = True

        # Create an array from all TR's with the inner HTML for each TR
        # Data we want is stored inside an arbitrary # of 'span' tags inside the TR's.\
        tr_list = soup.find_all("tr")

        # This will create an array for each TR with an array of SPAN values inside.
        feature_list = []
        for tag in tr_list:
            try:
                # Create an innerhtml array for all spans within a single TR
                tag = [j.string for j in tag.findAll("span")]
                if set(tuple(tag)) & set(self.features):
                    try:
                        # Save the spans inner HTML if its not a header label
                        tag = [i for i in tag if "(each" not in i.lower()]
                    except AttributeError:
                        continue
                    feature_list.append(tag)

            except IndexError:
                continue

        # feature_list is an array [tr] of arrays [spans]. we want this flattened.
        # [tr1span1KEY, tr1span1VALUE, tr1span2KEY, tr1span2VALUE, tr2span1KEY, tr2span1VALUE, ]
        try:
            # flatten multidimensional list
            feature_list = [
                item.replace(':', '')
                for sublist in feature_list for item in sublist
            ]

        except AttributeError:
            pass

        return feature_list

    def distribute(self, feature_list):

        # feature_list ~= [html][tr][spans].innterHTML
        # [tr1span1KEY, tr1span1VALUE, tr1span2KEY, tr1span2VALUE, tr2span1KEY, tr2span1VALUE, ]
        def __pair(list_type):

            # break up elements with n-tuples greater than 2
            def __raw_business(i): return any( x in feature_list[i:i + 2][0] for x in INTERNAL_FIELDS )
            def __feature_list(i): return feature_list[i:i + 2][0] in FEATURES
            condition = __raw_business if list_type else __feature_list

            # then convert list of tuples to dict for faster lookup
            return [ tuple(feature_list[i:i + 2]) for i in range(0, len(feature_list), 2) if condition(i) ]

        raw_business = __pair(1) # [(x1,y1),(x2,y2),(x3,y3)] => INTERNAL_FIELDS
        feature_list = dict(__pair(0)) # FEATURES
        filtered_business = []

        # Party_Type = 'property address' not 'plaintiff' or 'defendant'
        # Input exists 'Business or Org Name' and == an Address
        for label, value in enumerate(raw_business):
            try:
                party_type = value[1].upper()
                section = raw_business[label + 1][0].upper()
                flag1 = party_type == "PROPERTY ADDRESS"
                flag2 = section == "BUSINESS OR ORGANIZATION NAME"
                if flag1 and flag2: filtered_business.append(raw_business[label + 1])

            except IndexError:
                print("Party Type issue at Case", feature_list["Case Number"])

        scraped_features = []

        for address in filtered_business:

            str_address = filter_addr(str(address[-1]))

            temp_features = {
                key: value for key, value in feature_list.items()
                if key in ["Title", "Case Type", "Case Number", "Filing Date"]
            }


            if temp_features["Case Type"].upper() == "FORECLOSURE":
                temp_features["Case Type"] = "Mortgage"

            elif temp_features["Case Type"].upper() == \
                    "FORECLOSURE RIGHTS OF REDEMPTION" and self.maybe_tax:
                temp_features["Case Type"] = "Tax"

            else:
                # break out of the rest of the loop if case type is neither
                continue

            if 'Title' not in temp_features:
              # print('feature_list');
              # print(feature_list);
              # print('\n \n raw_business');
              # print(raw_business);
              continue

            # break up Title feature into Plaintiff and Defendant
            try:
                temp_features["Plaintiff"], temp_features["Defendant"] = \
                    TITLE_SPLIT_PAT.split(temp_features["Title"])

            except ValueError:
                temp_features["Plaintiff"], temp_features["Defendant"] = \
                    (", ")

            temp_features["Address"] = \
                str_address if str_address else address[-1]

            temp_features["Zip Code"] = ''.join(ZIP_PAT.findall(address[-1]))

            temp_features["Partial Cost"] = ''.join(
                MONEY_PAT.findall(address[-1])
            )

            scraped_features.append(temp_features)
            temp_features = {}

        return scraped_features


# Cell
# from settings import CHECKPOINT, HTML_DIR
from os import path, remove, walk

temp_output = "temp_data.csv"
file_array = [filenames for (dirpath, dirnames, filenames) in walk(HTML_DIR)][0]
top5 = file_array[:1]

# Cell
"""Cleaner

This module implements post-scraping cleaning processes on the raw initial
dataset. Processes include stripping excess strings off Address values,
removing Zip Code and Partial Cost values mislabeled as Address, and merging
rows containing blank values in alternating features.

The script works as an internal module for Close Crawl, but can be executed
as a standalone to manually process datasets:

    $ python cleaner.py <path/to/old/dataset> <path/of/new/dataset>

"""

# Cell
#nbdev_comment from __future__ import absolute_import, print_function, unicode_literals
from pandas import DataFrame, concat, read_csv, to_datetime
# from patterns import NULL_ADDR, STRIP_ADDR, filter_addr, punctuation

# Cell
class Cleaner(object):
    """Class object for cleaning the raw dataset extracted after the initial
    scraping
    """

    def __init__(self, path):
        """Constructor for Cleaner

        Args:
            path (`str`): path to input CSV dataset

        Attributes:
            df (`pandas.core.frame.DataFrame`): initial DataFrame
            columns (`list` of `str`): columns of the DataFrame
            clean_df (`pandas.core.frame.DataFrame`): final DataFrame to be
                outputted
        """

        self.df = self.prettify(read_csv(path))

        self.columns = list(self.df)
        self.clean_df = []

    @staticmethod
    def prettify(df, internal=True):
        """Drops duplicates, sorts and fills missing values in the DataFrame
        to make it manageable.

        Args:
            df (`pandas.core.frame.DataFrame`): DataFrame to be managed
            internal (`bool`, optional): flag for determining state of
                DataFrame

        Returns:
            df (`pandas.core.frame.DataFrame`): organized DataFrame
        """

        df.drop_duplicates(inplace=True, keep='last', subset=["Case Number"] )
        df["Filing Date"] = to_datetime(df["Filing Date"])

        df.sort_values(
            ["Filing Date", "Case Number", "Address"],
            ascending=[True] * 3,
            inplace=True
        )

        if internal:
            df["Zip Code"] = df["Zip Code"].fillna(0.0).astype(int)
            df["Zip Code"] = df["Zip Code"].replace(0, '')
        return df

    def clean_addr(self):
        """Cleans excess strings off Address values and removes Zip Code and
        Partial Cost values mislabeled as Address.

        Args:
            None

        Returns:
            None
        """

        def clean_string(addr):
            """Applies regular expressions and other filters on Address
            values

            Args:
                addr (`str`): Address value to be filtered

            Returns:
                addr (`str`): filtered Address value
            """

            # if value does not match the street_address pattern
            if not filter_addr(addr):  # patterns.filter_addr

                if NULL_ADDR.sub('', addr):  # value may contain valid Address
                    return str(
                        STRIP_ADDR.sub(
                            '', addr)  # strip off Zip Code and Partial Cost
                    ).translate(
                        {ord(c): None for c in punctuation}
                    ).strip()  # strip off punctuations

            return addr

        print("Cleaning addresses...", end=" ")

        self.df["Address"] = self.df["Address"].apply(
            lambda x: clean_string(x)
        )
        self.df["Address"] = self.df["Address"].apply(
            lambda x: NULL_ADDR.sub('', x)
        )

        # replace empty string values with NULL
        self.df["Zip Code"] = self.df["Zip Code"].replace('', float("nan"))
        self.df["Address"] = self.df["Address"].replace('', float("nan"))

        print("Done")

    @staticmethod
    def combine_rows(row):
        """Merges rows after filtering out common values

        Args:
            row (`list` of `list` of `str`): groupby("Case Number") rows

        Returns:
            (`list` of `str`): merged row
        """

        def __filter_tuple(col):
            """Filters common values from rows

            Args:
                col (`tuple` of `str`): values per column

            Returns:
                value (`str`): common value found per mergeable rows
            """

            for value in set(col):
                if value == value:  # equivalent to value != NaN
                    return value

        return [__filter_tuple(x) for x in zip(*row)]

    @staticmethod
    def mergeable(bool_vec):
        """Determines if groupby("Case Number") rows are mergeable

        Example:
            bool_vec = [
                [True, True, True, True, True, True, False, True, True],
                [True, True, True, True, True, True, True, False, False],
                [True, True, True, True, True, True, False, False, False]
            ]

            __sum_col(bool_vec) -> [3, 3, 3, 3, 3, 3, 1, 1, 1]

            __bool_pat(__sum_col(bool_vec)) -> True

        Args:
            bool_vec (`list` of `bool`): represents non-NULL values

        Returns:
            (`bool`): True if rows are mergeable
        """

        def __sum_col():
            """Sums columns

            Args:
                None

            Returns:
                (`list` of `int`): sum of columns
            """
            return [sum(x) for x in zip(*bool_vec)]

        def __bool_pat(row):
            """Determines mergeability

            Args:
                None

            Returns:
                (`bool`): True if rows are mergeable
            """
            return set(row[-3:]) == set([1]) and set(row[:-3]) != set([1])

        return True if __bool_pat(__sum_col()) else False

    def merge_nulls(self):
        """Splits DataFrames into those with NULL values to be merged, and then
        later merged with the original DataFrame

        Args:
            None

        Returns:
            None
        """

        print("Merging rows...", end=" ")
        print(self.df)

        # filter out rows with any NULL values
        origin_df = self.df.dropna()

        # filter out rows only with NULL values
        null_df = self.df[self.df.isnull().any(axis=1)]

        # boolean representation of the DataFrame with NULL values
        bool_df = null_df.notnull()

        # (`list` of `dict` of `str` : `str`) to be converted to a DataFrame
        new_df = []

        for i in null_df["Case Number"].unique():
            bool_row = bool_df[null_df["Case Number"] == i]
            new_row = null_df[null_df["Case Number"] == i]

            # if the rows are mergeable, combine them
            if self.mergeable(bool_row.values):
                new_row = self.combine_rows(new_row.values.tolist())

                new_df.append(
                    {
                        feature: value
                        for feature, value in zip(self.columns, new_row)
                    }
                )

            # else, treat them individually
            else:
                new_row = new_row.values.tolist()

                for row in new_row:
                    new_df.append(
                        {
                            feature: value
                            for feature, value in zip(self.columns, row)
                        }
                    )

        # merge the DataFrames back
        self.clean_df = concat(
            [origin_df, DataFrame(new_df)]
        ).reset_index(drop=True)

        # prettify the new DataFrame
        self.clean_df = self.prettify(
            self.clean_df[self.columns], internal=False
        )

        print("Done")

    def init_clean(self):
        """Initializes cleaning process

        Args:
            None

        Returns:
            None
        """
        print(self.df)
        self.clean_addr()
        print(self.df)
        self.merge_nulls()

    def download(self, output_name):
        """Downloads the cleaned and manipulated DataFrame into a CSV file

        Args:
            output_name (`str`): path of the new output file

        Returns:
            None
        """
        self.clean_df.to_csv(output_name, index=False)


# Cell
temp_output = "temp_data.csv"
output = 'outputfile.csv'
df_obj = Cleaner(temp_output)
df_obj.init_clean()
df_obj.download(output)

# Cell
"""main

The main executable script for Close Crawl. This file manages types, flags
and constraints for the case type, year and output data file.

Usage:
    $ python main.py <case_type> <case_year> <path/of/new/dataset>
      <opt: lower_bound> <opt: upper_bound> <opt: debug>

Example usage:
    $ python main.py O 2015 test_set.csv -l=300 -u=600 -d=1

"""

# Cell
#nbdev_comment from __future__ import absolute_import, print_function, unicode_literals
from json import dump, dumps, load
from os import path, remove, walk
from shutil import rmtree
from time import time

from cleaner import Cleaner
from miner import Miner
from settings import CHECKPOINT, HTML_DIR
from spider import Spider

# Cell
def close_crawl(case_type, case_year, output, cases='', lower_bound=0,
                upper_bound=0, debug=False, scrape=True, mine=True,
                clean=True):
    """Main function for Close Crawl.

    Args:
        case_type (`str`): type of foreclosure case, options are 'O' and 'C'
        case_year (`str`): year of foreclosure cases
        output (`str`): path of the output CSV file, along with the valid
            extension (.csv)
        lower_bound (`int`, optional): lower bound of range of cases
        upper_bound (`int`, optional): upper bound of range of cases
        debug (`bool`, optional): option for switching between debug mode.
            Default -> True

    Returns:
        None
    """

    temp_output = "temp_data.csv"
    wait = 0
    case_list = []

    print('checkpoint')
    if not path.isfile(CHECKPOINT):
        print("Initializing project...")
        with open(CHECKPOINT, "w") as checkpoint:
            dump(
                {
                    "last_case": "{:04d}".format(int(str(lower_bound)[-4:])),
                    "type": case_type,
                    "year": case_year[-2:],
                    "error_case": '',
                },
                checkpoint
            )

    print('cases')
    if not cases:

        with open(CHECKPOINT) as checkpoint:
            prev_bound = int(load(checkpoint)["last_case"])
            if not lower_bound:
                lower_bound = prev_bound
            upper_bound = upper_bound if int(upper_bound) > int(lower_bound) \
                else str(lower_bound + 5)

        case_list = range(int(lower_bound), int(upper_bound) + 1)

    else:

        with open(cases) as manual_cases:
            case_list = sorted(list(set(load(manual_cases))))

    print('scrape')
    if scrape:
        spider = Spider(
            case_type=case_type, year=case_year[-2:],
            bounds=case_list, gui=False
        )

        spider.save_response()

        wait = spider.WAITING_TIME

    file_array = [filenames for (dirpath, dirnames, filenames)
                  in walk(HTML_DIR)][0]

    start_mine = time()
    print('mine')
    if mine:
        miner = Miner(file_array, temp_output)
        miner.scan_files()
        miner.export()


    print('clean')
    if clean:
        df_obj = Cleaner(temp_output)

        df_obj.init_clean()
        df_obj.download(output)

    print('save')
    with open(CHECKPOINT, "r+") as checkpoint:
        checkpoint_data = load(checkpoint)
        checkpoint_data["last_case"] = sorted(file_array)[-1].split('.')[0][-4:]
        checkpoint.seek(0)
        checkpoint.write(dumps(checkpoint_data))
        checkpoint.truncate()


    print("Crawling runtime: {0:.2f} s".format((end_crawl - start_crawl)))
    print(
        "Downloading runtime: {0:.2f} s".format(
            ((end_crawl - start_crawl) - wait))
    )
    print("Mining runtime: {0:.2f} s".format((end_mine - start_mine)))
    print("Program runtime: {0:.2f} s".format((end - start)))
    print("------------ SCRAPING COMPLETED ------------")


# Cell
args = {'type': 'C', 'year': '2014', 'output': 'outputfile.csv', 'file': '', 'lower': '1000', 'upper': '2000', 'debug': '0', 'scrape': True, 'mine': True, 'clean': True}
close_crawl(
    case_type=args["type"], case_year=args["year"], output=args["output"],
    cases=args["file"], lower_bound=args["lower"],
    upper_bound=args["upper"], debug=args["debug"],
    scrape=args["scrape"], mine=args["mine"],
    clean=args["clean"]
)

# Cell
"""cliargs

The main command line script for Close Crawl. This file manages types, flags
and constraints for the case type, year and output data file as well as the
processing options.

Parameters:
  {O,C}        | Type of foreclosure cases
  year         | Year of foreclosure cases
  output       | Path of output file

Optional parameters:
  -h, --help   | Show this help message and exit
  -v, --version| Show program's version number and exit
  -l, --lower  | Lower bound of range of cases
  -u, --upper  | Upper bound of range of cases
  -f, --file   | Path of JSON array of cases
  -d, --debug  | Debug mode
  -s, --scrape | Scrape only
  -m, --mine   | Mine only
  -c, --clean  | Clean only

Usage:
    $ python cliarg.py [-h] [-v] [-l] [-u] [-f] [-d] [-s] [-m] [-c]
                  {O,C} year output

Example usages:
    $ python cliarg.py -l=50 -u=3500 -d -s -m C 2016 output.csv
    $ python cliarg.py -c="cases_to_scrape.json" -d O 2014 output01.csv
"""

# Cell
#nbdev_comment from __future__ import absolute_import, print_function, unicode_literals
import sys
from textwrap import dedent

from modules import main

# Cell
#!/usr/bin/env python
# -*- coding: utf-8 -*-

if __name__ == "__main__":

    args = {}

    args["type"] = input("Enter type of case (1 char: {C, O}): ")
    args["year"] = input("Enter year of case (4 digit int): ")
    args["output"] = input("Enter name of output file (CSV file path): ")

    opt = int(input("Enter 0 for manual parameters or 1 for automatic: "))
    if bool(opt):
        args["file"] = input("Enter name of cases file (JSON file path): ")
        args["lower"] = args["upper"] = 0

    else:
        args["file"] = ""
        args["lower"] = input("Enter lower bound of cases (1-4 digit int): ")
        args["upper"] = input("Enter upper bound of cases (1-4 digit int): ")

    args["debug"] = input(
        "Enter 0 for default mode, 1 for debug (1 digit int): "
    )

    print(
        dedent(
            """Processing options:\n\n"""
            """For the following options, enter 0 to disable or 1 to enable."""
            """\nNOTE: The script will exit if all but the mining step is """
            """enabled - data cannot be cleaned without being mined first."""
        )
    )

    args["scrape"] = bool(input("Scrape: {0, 1}: "))
    args["mine"] = bool(input("Mine: {0, 1}: "))
    args["clean"] = bool(input("Clean: {0, 1}: "))

    # exit script if all but the mining step is enabled
    if (args["scrape"] and not(args["mine"]) and args["clean"]):
        sys.exit("\nData cannot be cleaned without being mined first.")

    print (args)
    main.close_crawl(
        case_type=args["type"], case_year=args["year"], output=args["output"],
        cases=args["file"], lower_bound=args["lower"],
        upper_bound=args["upper"], debug=args["debug"],
        scrape=args["scrape"], mine=args["mine"],
        clean=args["clean"]
    )



# AUTOGENERATED! DO NOT EDIT! File to edit: notebooks/02_TidyAddr.ipynb (unless otherwise specified).

__all__ = ['runTidyAddr']

# Cell
# Input: A CSV with a single column titled 'address' (may require pre-processing to get it like this)
# Output: A CSV with a single column containing text that can be split into 5 columns in xl.
# - Needs to be re-merged with any other columns removed before tidy addring.
import os
def runTidyAddr( filename, newfilename ):
  print( filename, newfilename )
  print(' ~~~~~~~~~~~ Getting TIDYADDR ~~~~~~~~~~~ ')
  ! rm tidyaddr-js -r
  ! git clone https://github.com/BNIA/tidyaddr-js.git
  print(' ~~~~~~~~~~~ Installing TIDYADDR ~~~~~~~~~~~')
  ! echo ORIGINAL_DIRECTORY && echo $(ls)
  ! cp *.csv tidyaddr-js/ && cd tidyaddr-js && npm install
  print(' ~~~~~~~~~~~ Running TIDYADDR ~~~~~~~~~~~')
  ! echo TIDYADDR_DIRECTORY && cd tidyaddr-js && echo $(ls)
  txt = "cd tidyaddr-js && node tidyaddr.js clean-csv " + filename + " " + newfilename
  os.system(txt)

runTidyAddr( uploadedFilename, './'+uploadedFilename.replace(".", "_tidyaddred.") )


# AUTOGENERATED! DO NOT EDIT! File to edit: notebooks/CitiStat_Dirtyst_Clogged.ipynb (unless otherwise specified).

__all__ = ['dirtyst', 'dirtyst', 'dirtyst', 'dirtyst', 'dirtyst', 'clogged', 'clogged', 'clogged', 'clogged', 'clogged',
           'lights', 'fin']

# Cell

# Copy the Data
dirtyst = dirtydf.copy()

# Query the Data
# dirtyst = dirtyst[ (dirtyst['SRType'].str.contains('SW-Dirty Alley|SW-Dirty Street', regex=True) ) ]

# *Special*: # UPDATE HERE AND THEN GROUP
dirtyst['162-dirtyst'+year] = 1
dirtyst = dirtyst.append({'CSA2010': 'Baltimore City' , '162-dirtyst'+year : dirtyst['162-dirtyst'+year].sum() } , ignore_index=True)
dirtyst = dirtyst.groupby('CSA2010').sum(numeric_only=True)

# Make sure ALL csas and BaltimoreCity are included and sorted.
dirtyst = csa.merge( dirtyst, left_on='CSA2010', right_on='CSA2010', how='outer' )
dirtyst.drop(columns=['Shape__Length', 'Shape__Area', 'geometry' ], inplace=True)
dirtyst = dirtyst[['CSA2010', '162-dirtyst'+year, 'tpop10']]


# Create the Indicator
dirtyst['162-dirtyst'+year] = dirtyst['162-dirtyst'+year] * 1000 / dirtyst['tpop10']

display( dirtyst.head(2) )
dirtyst.tail(2)

dirtyst.to_csv('162-dirtyst'+year+'.csv')

# Cell

# Copy the Data
clogged = cloggedf.copy()

# Query the Data
# clogged = clogged[
#  (dirtyst['PROPERTYIN'].str.contains('CONDOMINIUM|SINGLE FAMILY', regex=True) )
#| (dirtyst['SALEDATE'] >= '20'+year+'-01-01')
#]

# *Special*: # UPDATE HERE AND THEN GROUP
clogged['163-clogged'+year] = 1
clogged = clogged.append({'CSA2010': 'Baltimore City' , '163-clogged'+year : clogged['163-clogged'+year].sum() } , ignore_index=True)
clogged = clogged.groupby('CSA2010').sum(numeric_only=True)

# Make sure ALL csas and BaltimoreCity are included and sorted.
clogged = csa.merge( clogged, left_on='CSA2010', right_on='CSA2010', how='outer' )
clogged.drop(columns=['Shape__Length', 'Shape__Area', 'geometry' ], inplace=True)
clogged = clogged[['CSA2010', '163-clogged'+year, 'tpop10']]


# Create the Indicator
clogged['163-clogged'+year] = clogged['163-clogged'+year] * 1000 / clogged['tpop10']

display( clogged.head(2) )
clogged.tail(2)

clogged.to_csv('163-clogged'+year+'.csv')

# Cell
def lights(df, csa, yr):
  # The rate of service requests for addressing street light outages made through Baltimore's 311 system per 1,000 residents.
  # More than one service request may be made for the same issue but is logged as a unique request.

  # Create the Numerator
  lights = df.copy()
  lights['count'] = 1
  lights = lights.groupby('CSA2010').sum(numeric_only=True)

  # Make sure ALL csas and BaltimoreCity are included and sorted.
  lights = csa.merge( lights, left_on='CSA2010', right_on='CSA2010', how='outer' )
  lights.drop( columns=['geometry', 'Shape__Length','CouncilDis','Latitude','Longitude','Shape__Area','OBJECTID_y','OBJECTID_x'], inplace=True)
  # Baltimoire has records not in the
  lights.at[55,'count']=lights['count'].sum()
  # Perform the calculation
  lights['215-lights'+year] = lights['count'] / lights['tpop10'] * 1000

  compareYears = gpd.read_file("https://services1.arcgis.com/mVFRs7NF4iFitgbY/ArcGIS/rest/services/Lights/FeatureServer/0/query?where=1%3D1&outFields=*&returnGeometry=true&f=pgeojson");
  prevYear = 'lights'+ str( int(year) - 1 )
  if prevYear in compareYears.columns:
    lights = lights.merge( compareYears[['CSA2010', prevYear]], left_on='CSA2010', right_on='CSA2010', how='outer' )
    lights['change'] = lights['215-lights'+year] - lights[ prevYear ]
    lights['percentChange'] = lights['change' ] / lights[ prevYear ] * 100
    lights['change'] = lights['change'].apply(lambda x: "{:.2f}".format(x) )
  print( 'Records Matching Query: ', lights.size / len(lights.columns) )
  return lights

fin = lights(df, csa, year)
fin.to_csv('215-lights'+year+'.csv', index=False)
fin.head(60)


# AUTOGENERATED! DO NOT EDIT! File to edit: notebooks/CityFinance_histaxhometaxowntax.ipynb (unless otherwise specified).

__all__ = ['homtax']

# Cell
def homtax(df, totalres, year):
  homtax = df.copy()
  # Aggregate Numeric Values by Sum
  homtax['homtaxCount'] = 1
  homtax = homtax.groupby('CSA2010').sum(numeric_only=True)

  # Make sure ALL csas and BaltimoreCity are included. among other things
  homtax = totalres[ ['CSA2010','totalres'+year] ].merge( homtax, left_on='CSA2010', right_on='CSA2010', how='outer' )

  # Update the baltimore CSA.
  homtax.at[55,'homtaxCount'] = homtax['homtaxCount'].sum()

  # Create the Indicator
  homtax['47-homtax'+year] = homtax['homtaxCount'] * 1000 / totalres['totalres'+year]

  homtax = homtax[['CSA2010', 'homtaxCount', '47-homtax'+year, 'totalres'+year ]]

  compareYears = gpd.read_file("https://services1.arcgis.com/mVFRs7NF4iFitgbY/ArcGIS/rest/services/Histax/FeatureServer/0/query?where=1%3D1&outFields=*&returnGeometry=true&f=pgeojson");
  prevYear = 'histax'+ str( int(year) - 1 )
  if prevYear in compareYears.columns:
    homtax = homtax.merge( compareYears[['CSA2010', prevYear]], left_on='CSA2010', right_on='CSA2010', how='outer' )
    homtax['change'] = homtax['47-homtax'+year] - homtax[ prevYear ]
    homtax['percentChange'] = homtax['change'] / homtax[ prevYear ] * 1000
    homtax['change'] = homtax['change'].apply(lambda x: "{:.2f}".format(x) )
    homtax = homtax[['CSA2010', 'homtaxCount', 'totalres19', 'homtax18', '47-homtax19', 'percentChange', 'change']]

  return homtax


# AUTOGENERATED! DO NOT EDIT! File to edit: notebooks/BPD_P1.ipynb (unless otherwise specified).

__all__ = []

# Cell
dft['crime'+year] = (
    dft['Desc__AGG. ASSAULT'] +
    dft['Desc__AUTO THEFT'] +
    dft['Desc__BURGLARY'] +
    dft['Desc__HOMICIDE'] +
    dft['Desc__LARCENY'] +
    dft['Desc__LARCENY FROM AUTO'] +
    dft['Desc__RAPE'] +
    dft['Desc__ROBBERY - CARJACKING'] +
    dft['Desc__ROBBERY - COMMERCIAL'] +
    dft['Desc__ROBBERY - RESIDENCE'] +
    dft['Desc__ROBBERY - STREET']
) * 1000 / csa['tpop10']

# Just checking I got all but the three. 14 desc cols - 11 being used = 3 not used. Alls good.

# Cell
dft['viol'+year] = (dft['Desc__ROBBERY - CARJACKING'] +
                   dft['Desc__ROBBERY - COMMERCIAL'] +
                   dft['Desc__ROBBERY - RESIDENCE'] +
                   dft['Desc__ROBBERY - STREET'] +
                   dft['Desc__AGG. ASSAULT'] +
                   dft['Desc__RAPE'] +
                   dft['Desc__HOMICIDE'] ) * 1000 / csa['tpop10']

# Cell
dft['prop'+year] = (dft['Desc__LARCENY'] +
                   dft['Desc__LARCENY FROM AUTO'] +
                   dft['Desc__BURGLARY'] +
                   dft['Desc__AUTO THEFT'] ) * 1000 / csa['tpop10']

# Cell
dft['gunhom'+year] = (dft['gunhom'] ) * 1000 / csa['tpop10']


# AUTOGENERATED! DO NOT EDIT! File to edit: notebooks/05_Create_Acs_Indicators.ipynb (unless otherwise specified).

__all__ = ['getColName', 'getColByName', 'addKey', 'nullIfEqual', 'sumInts', 'mhhi', 'trav45', 'trav44', 'age5',
           'age65', 'age18', 'paa', 'hisp', 'pwhite', 'hh25inc', 'novhcl', 'nohhint', 'createAcsIndicator']

# Cell

#@title Run This Cell: Misc Function Declarations
# These functions right here are used in the calculations below.
# Finds a column matchings a substring
def getColName  (df, col): return df.columns[df.columns.str.contains(pat = col)][0]
def getColByName (df, col): return df[getColName(df, col)]

# Pulls a column from one dataset into a new dataset.
# This is not a crosswalk. calls getColByName()
def addKey(df, fi, col):
    key = getColName(df, col)
    val = getColByName(df, col)
    fi[key] = val
    return fi
# Return 0 if two specified columns are equal.
def nullIfEqual(df, c1, c2):
    return df.apply(lambda x:
        x[getColName(df, c1)]+x[getColName(df, c2)] if x[getColName(df, c1)]+x[getColName(df, c2)] != 0 else 0, axis=1)
# I'm thinking this doesnt need to be a function..
def sumInts(df): return df.sum(numeric_only=True)

# Cell
# @title Run This Cell : Create MHHI

#File: mhhi.py
#Author: Charles Karpati
#Date: 1/24/19
#Section: Bnia
#Email: karpati1@umbc.edu
#Description:
# Uses ACS Table B19001 - HOUSEHOLD INCOME IN THE PAST 12 MONTHS (IN 2016 INFLATION-ADJUSTED DOLLARS)
# Universe: Households
# Table Creates: hh25 hh40 hh60 hh75 hhm75, mhhi
#purpose: Produce Sustainability - Percent of Population that Walks to Work Indicator
#input:
#output:
import pandas as pd
import glob

def mhhi( df, columnsToInclude = [] ):
  #~~~~~~~~~~~~~~~
  # Step 2)
  # Prepare the columns
  #~~~~~~~~~~~~~~~
  info = pd.DataFrame(
      [
          ['B19001_002E', 0, 10000],
          ['B19001_003E', 10000, 4999 ],
          ['B19001_004E', 15000, 4999 ],
          ['B19001_005E', 20000, 4999 ],
          ['B19001_006E', 25000, 4999 ],
          ['B19001_007E', 30000, 4999],
          ['B19001_008E', 35000, 4999 ],
          ['B19001_009E', 40000, 4999 ],
          ['B19001_010E', 45000, 4999 ],
          ['B19001_011E', 50000, 9999 ],
          ['B19001_012E', 60000, 14999],
          ['B19001_013E', 75000, 24999 ],
          ['B19001_014E', 100000, 24999 ],
          ['B19001_015E', 125000, 24999 ],
          ['B19001_016E', 150000, 49000 ],
          ['B19001_017E', 200000, 1000000000000000000000000 ],

      ],
      columns=['variable', 'lower', 'range']
  )

  # Final Dataframe
  data_table = pd.DataFrame()
  for index, row in info.iterrows():
      data_table = addKey(df, data_table, row['variable'])

  # Accumulate totals accross the columns.
  # Midpoint: Divide column index 16 (the last column) of the cumulative totals
  temp_table = data_table.cumsum(axis=1)
  temp_table['midpoint'] = (temp_table.iloc[ : , -1 :] /2) # V3
  temp_table['midpoint_index'] = False
  temp_table['midpoint_index_value'] = False # Z3
  temp_table['midpoint_index_lower'] = False # W3
  temp_table['midpoint_index_range'] = False # X3
  temp_table['midpoint_index_minus_one_cumulative_sum'] = False #Y3
  # step 3 - csa_agg3: get the midpoint index by "when midpoint > agg[1] and midpoint <= agg[2] then 2"

  # Get CSA Midpoint Index using the breakpoints in our info table.
  for index, row in temp_table.iterrows():
      # Get the index of the first column where our midpoint is greater than the columns value.
      midpoint = row['midpoint']
      midpoint_index = 0
      # For each column (except the 6 columns we just created)

      # The tracts midpoint was < than the first tracts value at column 'B19001_002E_Total_Less_than_$10,000'
      if( midpoint < int(row[0]) or row[-6] == False ):
        temp_table.loc[ index, 'midpoint_index' ] = 0
      else:
        for column in row.iloc[:-6]:
            # set midpoint index to the column with the highest value possible that is under midpoint
            if( midpoint >= int(column) ):
                if midpoint==False: print (str(column) + ' - ' + str(midpoint))
                temp_table.loc[ index, 'midpoint_index' ] = midpoint_index +1
            midpoint_index += 1

  # temp_table = temp_table.drop('Unassigned--Jail')
  for index, row in temp_table.iterrows():
    temp_table.loc[ index, 'midpoint_index_value' ] = data_table.loc[ index, data_table.columns[row['midpoint_index']] ]
    temp_table.loc[ index, 'midpoint_index_lower' ] = info.loc[ row['midpoint_index'] ]['lower']
    temp_table.loc[ index, 'midpoint_index_range' ] = info.loc[ row['midpoint_index'] ]['range']
    temp_table.loc[ index, 'midpoint_index_minus_one_cumulative_sum'] = row[ row['midpoint_index']-1 ]

  # This is our denominator, which cant be negative.
  for index, row in temp_table.iterrows():
    if row['midpoint_index_value']==False:
      temp_table.at[index, 'midpoint_index_value']=1;

  #~~~~~~~~~~~~~~~
  # Step 3)
  # Run the Calculation
  # Calculation = (midpoint_lower::numeric + (midpoint_range::numeric * ( (midpoint - midpoint_upto_agg) / nullif(midpoint_total,0)
  # Calculation = W3+X3*((V3-Y3)/Z3)

  # v3 -> 1 - midpoint of households  == sum / 2
  # w3 -> 2 - lower limit of the income range containing the midpoint of the housing total == row[lower]
  # x3 -> width of the interval containing the medium == row[range]
  # z3 -> number of hhs within the interval containing the median == row[total]
  # y3 -> 4 - cumulative frequency up to, but no==NOT including the median interval
  #~~~~~~~~~~~~~~~

  def finalCalc(x):
    return ( x['midpoint_index_lower']+ x['midpoint_index_range']*(
      ( x['midpoint']-x['midpoint_index_minus_one_cumulative_sum'])/ x['midpoint_index_value'] )
    )

  temp_table['final'] = temp_table.apply(lambda x: finalCalc(x), axis=1)


  columnsToInclude.append('tract')
  print ('INCLUDING COLUMN(s):' + str(columnsToInclude))
  temp_table[columnsToInclude] = df[columnsToInclude]
  #~~~~~~~~~~~~~~~
  # Step 4)
  # Add Special Baltimore City Data
  #~~~~~~~~~~~~~~~
  # url = 'https://api.census.gov/data/20'+str(year)+'/acs/acs5/subject?get=NAME,S1901_C01_012E&for=county%3A510&in=state%3A24&key=829bf6f2e037372acbba32ba5731647c5127fdb0'
  # table = pd.read_json(url, orient='records')
  # temp_table['final']['Baltimore City'] = float(table.loc[1, table.columns[1]])

  return temp_table

# Cell
#@title Run This Cell: Create trav45

#File: trav45.py
#Author: Charles Karpati
#Date: 1/17/19
#Section: Bnia
#Email: karpati1@umbc.edu
#Description:
# Uses ACS Table B08303 - TRAVEL TIME TO WORK,
# (Universe: Workers 16 years and over who did not work at home)
# Table Creates: trav14, trav29, trav44, trav45
#purpose: Produce Sustainability - Percent of Employed Population with Travel Time to Work of 45 Minutes and Over Indicator
#input:
#output:

import pandas as pd
import glob
def trav45(df, columnsToInclude = [] ):
    #~~~~~~~~~~~~~~~
    # Step 2)
    # Prepare the columns
    #~~~~~~~~~~~~~~~

    # Final Dataframe
    fi = pd.DataFrame()
    columns = ['B08303_011E','B08303_012E','B08303_013E','B08303_001E', 'tract']
    columns.extend(columnsToInclude)
    for col in columns:
        fi = addKey(df, fi, col)

    # Numerators
    numerators = pd.DataFrame()
    columns = ['B08303_011E','B08303_012E','B08303_013E']
    for col in columns:
        numerators = addKey(df, numerators, col)

    # Denominators
    denominators = pd.DataFrame()
    columns = ['B08303_001E']
    for col in columns:
        denominators = addKey(df, denominators, col)
    # construct the denominator, returns 0 iff the other two rows are equal.

    #~~~~~~~~~~~~~~~
    # Step 3)
    # Run the Calculation
# ( (value[1] + value[2] + value[3] ) / nullif(value[4],0) )*100
    #~~~~~~~~~~~~~~~
    fi['numerator'] = numerators.sum(axis=1)
    fi['denominator'] = denominators.sum(axis=1)
    fi = fi[fi['denominator'] != 0] # Delete Rows where the 'denominator' column is 0
    fi['final'] = (fi['numerator'] / fi['denominator'] ) * 100

    return fi

# Cell
#@title Run This Cell: Create trav44

#File: trav44.py
#Author: Charles Karpati
#Date: 1/17/19
#Section: Bnia
#Email: karpati1@umbc.edu
#Description:
# Uses ACS Table B08303 - TRAVEL TIME TO WORK,
# (Universe: Workers 16 years and over who did not work at home)
# Table Creates: trav14, trav29, trav44, trav45
#purpose: Produce Sustainability - Percent of Employed Population with Travel Time to Work of 30-44 Minutes Indicator
#input:
#output:

import pandas as pd
import glob
def trav44( df, columnsToInclude = [] ):
    #~~~~~~~~~~~~~~~
    # Step 2)
    # Prepare the columns
    #~~~~~~~~~~~~~~~

    # Final Dataframe
    fi = pd.DataFrame()
    columns = ['B08303_008E','B08303_009E','B08303_010E','B08303_001E', 'tract']
    columns.extend(columnsToInclude)
    for col in columns:
        fi = addKey(df, fi, col)

    # Numerators
    numerators = pd.DataFrame()
    columns = ['B08303_008E','B08303_009E','B08303_010E']
    for col in columns:
        numerators = addKey(df, numerators, col)

    # Denominators
    denominators = pd.DataFrame()
    columns = ['B08303_001E']
    for col in columns:
        denominators = addKey(df, denominators, col)
    # construct the denominator, returns 0 iff the other two rows are equal.

    #~~~~~~~~~~~~~~~
    # Step 3)
    # Run the Calculation
    # ( (value[1] + value[2] + value[3] ) / nullif(value[4],0) )*100
    #~~~~~~~~~~~~~~~
    fi['numerator'] = numerators.sum(axis=1)
    fi['denominator'] = denominators.sum(axis=1)
    fi = fi[fi['denominator'] != 0] # Delete Rows where the 'denominator' column is 0
    fi['final'] = (fi['numerator'] / fi['denominator'] ) * 100

    return fi


# Cell
#@title Run This Cell: Create age5

#File: age5.py
#Author: Charles Karpati
#Date: 4/16/19
#Section: Bnia
#Email: karpati1@umbc.edu
#Description:
# Uses ACS Table B01001 - SEX BY AGE
# Universe: Total population
# Table Creates: tpop, female, male, age5 age18 age24 age64 age65
#purpose:
#input: #output:

import pandas as pd
import glob
def age5( df, columnsToInclude ):

  #~~~~~~~~~~~~~~~
    # Step 1)
    # Prepare the columns
    #~~~~~~~~~~~~~~~
    # Final Dataframe
    fi = pd.DataFrame()
    columns = ['B01001_027E_Total_Female_Under_5_years',
               'B01001_003E_Total_Male_Under_5_years',
               'B01001_001E_Total' , 'tract']
    columns.extend(columnsToInclude)
    for col in columns:
        fi = addKey(df, fi, col)

    # Under 5
    fi['final']  = ( df[ 'B01001_003E_Total_Male_Under_5_years' ]
               + df[ 'B01001_027E_Total_Female_Under_5_years' ]
    ) / df['B01001_001E_Total'] * 100

    return fi

# Cell
#@title Run This Cell: age65

import pandas as pd
import glob
def age65( df, columnsToInclude ):
  #~~~~~~~~~~~~~~~
  # Step 1)
  # Prepare the columns
  #~~~~~~~~~~~~~~~
  # Final Dataframe
  fi = pd.DataFrame()

  columns = df.filter(regex='001E|020E|021E|022E|023E|024E|025E|044E|045E|046E|047E|048E|049E').columns.values
  columns = numpy.append(columns, columnsToInclude)

  for col in columns:
      fi = addKey(df, fi, col)

#  print('COLS ')
#  print(df.columns)
  print(' ')
  # over 65
  fi['age65']  = ( df.filter(regex='020E|021E|022E|023E|024E|025E|044E|045E|046E|047E|048E|049E').sum(axis=1)
) / df['B01001_001E_Total:'] * 100

  return fi

# Cell
#@title Run This Cell: age18

#File: age18.py
#Author: Charles Karpati
#Date: 4/16/19
#Section: Bnia
#Email: karpati1@umbc.edu
#Description:
# Uses ACS Table B01001 - SEX BY AGE
# Universe: Total population
# Table Creates: tpop, female, male, age5 age18 age24 age64 age65
#purpose:
#input: #output:

import pandas as pd
import glob
def age18( df, columnsToInclude ):
  print(df.columns)
  #~~~~~~~~~~~~~~~
  # Step 1)
  # Prepare the columns
  #~~~~~~~~~~~~~~~
  # Final Dataframe
  fi = pd.DataFrame()

  columns = ['B01001_001E_Total',
            'B01001_004E_Total_Male_5_to_9_years',
            'B01001_005E_Total_Male_10_to_14_years' ,
            'B01001_006E_Total_Male_15_to_17_years',
            'B01001_028E_Total_Female_5_to_9_years',
            'B01001_029E_Total_Female_10_to_14_years' ,
             'B01001_030E_Total_Female_15_to_17_years']
  columns = df.filter(regex='001E|004E|005E|006E|028E|029E|030E').columns.values
  columns = numpy.append(columns, columnsToInclude)
  for col in columns:
      fi = addKey(df, fi, col)
      print(' ')

  # Under 5
  fi['age18']  = ( df.filter(regex='004E|005E|006E|028E|029E|030E').sum(axis=1)
) / df['B01001_001E_Total:'] * 100
  return fi

# Cell
#@title Run This Cell: paa

import pandas as pd
import glob
def paa( df, columnsToInclude ):
  print(df.columns)
  #~~~~~~~~~~~~~~~
  # Step 1)
  # Prepare the columns
  #~~~~~~~~~~~~~~~
  # Final Dataframe
  fi = pd.DataFrame()

  columns = ['B03002_001E_Total:',
             'B03002_004E_Total_Not_Hispanic_or_Latino_Black_or_African_American_alone']
  columns = df.filter(regex='001E|004E').columns.values
  columns = numpy.append(columns, columnsToInclude)

  for col in columns:
      print('addKey df',df.columns,'fi',fi.columns,'col: ', col)
      fi = addKey(df, fi, col)
      print(' ')

  # Calculate
  fi['paa']  = ( df.filter(regex='004E').sum(axis=1)
) / df['B03002_001E_Total:'] * 100

  return fi

# Cell
#@title Run This Cell:  hisp

import pandas as pd
import glob
def hisp( df, columnsToInclude ):
  print(df.columns)
  #~~~~~~~~~~~~~~~
  # Step 1)
  # Prepare the columns
  #~~~~~~~~~~~~~~~
  # Final Dataframe
  fi = pd.DataFrame()

  columns = ['B03002_001E_Total',
             'B03002_012E_Total_Hispanic_or_Latino']
  columns = df.filter(regex='001E|012E').columns.values
  columns = numpy.append(columns, columnsToInclude)
  for col in columns:
      print('addKey df',df.columns,'fi',fi.columns,'col: ', col)
      fi = addKey(df, fi, col)
      print(' ')

  # Calculate
  fi['hisp']  = ( df.filter(regex='012E').sum(axis=1)
) / df['B03002_001E_Total:'] * 100

  return fi

# Cell
#@title Run This Cell: pwhite

import pandas as pd
import glob
def pwhite( df, columnsToInclude ):
  print(df.columns)
  #~~~~~~~~~~~~~~~
  # Step 1)
  # Prepare the columns
  #~~~~~~~~~~~~~~~
  # Final Dataframe
  fi = pd.DataFrame()
  columns = ['B03002_001E_Total',
             'B03002_003E_Total_Not_Hispanic_or_Latino_White_alone']
  columns = df.filter(regex='001E|003E').columns.values
  columns = numpy.append(columns, columnsToInclude)
  for col in columns:
      print('addKey df',df.columns,'fi',fi.columns,'col: ', col)
      fi = addKey(df, fi, col)
      print(' ')

  # Calculate
  fi['pwhite']  = ( df.filter(regex='003E').sum(axis=1)
) / df['B03002_001E_Total:'] * 100

  return fi

# Cell
#@title Run This Cell: hh25inc

import pandas as pd
import glob
def hh25inc( df, columnsToInclude ):
  # df.columns = df.columns.str.replace(r"[,]", "")
  df.columns = df.columns.str.replace(r"[$]", "")
  print(df.columns)
  #~~~~~~~~~~~~~~~
  # Step 1)
  # Prepare the columns
  #~~~~~~~~~~~~~~~
  # Final Dataframe
  fi = pd.DataFrame()
  columns = ['B19001_001E_Total',
       "B19001_002E_Total_Less_than_10,000",
       "B19001_003E_Total_10,000_to_14,999",
       "B19001_004E_Total_15,000_to_19,999",
       "B19001_005E_Total_20,000_to_24,999"]
  columns = df.filter(regex='002E|003E|004E|005E|001E').columns.values
  columns = numpy.append(columns, columnsToInclude)
  for col in columns:
      print('addKey col: ', col, df.columns)
      fi = addKey(df, fi, col)

  # Calculate
  fi['hh25inc']  = ( df.filter(regex='002E|003E|004E|005E').sum(axis=1)
) / df['B19001_001E_Total:'] * 100

  return fi

# Cell
#@title Run This Cell: novhcl

import pandas as pd
import glob
def novhcl( df, columnsToInclude ):
  print(df.columns)
  #~~~~~~~~~~~~~~~
  # Step 1)
  # Prepare the columns
  #~~~~~~~~~~~~~~~
  # Final Dataframe
  fi = pd.DataFrame()

  columns = ['B08201_002E_Total_No_vehicle_available','B08201_001E_Total']
  columns = df.filter(regex='002E|003E|004E|005E|001E').columns.values
  columns = numpy.append(columns, columnsToInclude)
  for col in columns:
      print('addKey df',df.columns,'fi',fi.columns,'col: ', col)
      fi = addKey(df, fi, col)
      print(' ')

  # Calculate
  fi['novhcl']  = ( df.filter(regex='002E').sum(axis=1)
) / df['B08201_001E_Total:'] * 100

  return fi

# Cell
#@title Run This Cell: nohhint

import pandas as pd
import glob
def nohhint( df, columnsToInclude ):
  print(df.columns)
  #~~~~~~~~~~~~~~~
  # Step 1)
  # Prepare the columns
  #~~~~~~~~~~~~~~~
  # Final Dataframe
  fi = pd.DataFrame()

  columns = ['B28011_001E_Total',
       'B28011_002E_Total_With_an_Internet_subscription',
       'B28011_003E_Total_With_an_Internet_subscription_Dial-up_alone',
       'B28011_004E_Total_With_an_Internet_subscription_Broadband_such_as_cable,_fiber_optic,_or_DSL',
       'B28011_005E_Total_With_an_Internet_subscription_Satellite_Internet_service',
       'B28011_006E_Total_With_an_Internet_subscription_Other_service',
       'B28011_007E_Total_Internet_access_without_a_subscription',
       'B28011_008E_Total_No_Internet_access']
  columns = df.filter(regex='008E|001E').columns.values
  columns = numpy.append(columns, columnsToInclude)
  for col in columns:
      print('addKey df',df.columns,'col: ', col)
      fi = addKey(df, fi, col)
      print(' ')

  # Calculate
  fi['nohhint']  = ( df.filter(regex='008E').sum(axis=1)
) / df['B28011_001E_Total:'] * 100

  return fi

# Cell
#@title Run This Cell: Create createIndicator()
import geopandas as gpd
import numpy as np
import pandas as pd
from .acsDownload import retrieve_acs_data
from dataplay.merge import mergeDatasets
from dataplay.intaker import Intake
def createAcsIndicator(state, county, tract, year, tableId, saveAcs,
                    mergeUrl, merge_left_col, merge_right_col, merge_how, groupBy,
                    aggMethod, method, columnsToInclude, finalFileName=False):

  # Pull the data
  df = retrieve_acs_data(state, county, tract, tableId, year, saveAcs)
  print('Table: ' + tableId + ', Year: ' + year + ' imported.')

  # Get the crosswalk
  if mergeUrl:
    right_ds = Intake.getData( mergeUrl )
    print( right_ds.columns )
    print('Merge file imported')
    # Merge crosswalk with the data
    df = mergeDatasets( left_ds=df, right_ds=right_ds,
                  left_col=merge_left_col, right_col=merge_right_col,
                  merge_how=merge_how, interactive=False )

    print('Both are now merged.')

  # Group and Aggregate
  if groupBy:
    df = df.groupby(groupBy)
    print('Aggregating...')
    if aggMethod == 'sum':
      df = sumInts(df)
    else:
      df = sumInts(df)
    print('Aggregated')

  # Create the indicator
  print('Creating Indicator')
  resp = method( df, columnsToInclude)
  print('Indicator Created')
  if finalFileName:
    resp.to_csv(finalFileName, quoting=csv.QUOTE_ALL)
    print('Indicator Saved')

  return resp


# AUTOGENERATED! DO NOT EDIT! File to edit: notebooks/90_ACS_Explore_and_Download.ipynb (unless otherwise specified).

__all__ = ['retrieve_acs_data']

# Cell
# @title Run: Import Modules

# Once installed we need to..

# import and configure the Widgets
import ipywidgets as widgets
from IPython.core.interactiveshell import InteractiveShell
InteractiveShell.ast_node_interactivity = 'all'
import ipywidgets as widgets
from ipywidgets import interact, interact_manual

# About importing data
import urllib.request as urllib
from urllib.parse import urlencode
# This Prevents Timeouts when Importing
import socket
socket.setdefaulttimeout(10.0)

# Pandas Data Manipulation Libraries
import pandas as pd
# Working with Json Data
import json
# Data Processing
import numpy as np
# Reading Json Data into Pandas
from pandas.io.json import json_normalize

# Export data as CSV
import csv


# Geo-Formatting
# Postgres-Conversion
import geopandas as gpd
from geopandas import GeoDataFrame
import psycopg2,pandas,numpy
from shapely import wkb
from shapely.wkt import loads
import os
import sys

# In case file is KML
# enable KML support; disabled by default
import fiona
fiona.drvsupport.supported_drivers['kml'] = 'rw'
fiona.drvsupport.supported_drivers['KML'] = 'rw'

# load libraries
# from shapely.wkt import loads
# from pandas import ExcelWriter
# from pandas import ExcelFile
import matplotlib.pyplot as plt
import glob
import imageio

# Cell
# @title Run: Create retrieve_acs_data()

#File: retrieveAcsData.py
#Author: Charles Karpati
#Date: 1/9/19
#Section: Bnia
#Email: karpati1@umbc.edu
#Description:
#This file returns ACS data given an ID and Year
# The county total is given a tract of '010000'

#def retrieve_acs_data():
#purpose: Retrieves ACS data from the web
#input:
# state (required)
# county (required)
# tract (required)
# tableId (required)
# year (required)
# includeCountyAgg (True)(todo)
# replaceColumnNames (False)(todo)
# save (required)
#output:
# Acs Data.
# Prints to ../../data/2_cleaned/acs/

def retrieve_acs_data(state, county, tract, tableId, year, save):
    dictionary = ''
    keys = []
    vals = []
    header = []
    keys1=keys2=keys3=keys4=keys5=keys6=keys7=keys8=''
    keyCount = 0

    # Called in addKeys(), Will create the final URL for readIn()
    # These are parameters used in the API URL Query
    # This query will retrieve the census tracts
    def getParams(keys): return {
        'get': 'NAME'+keys,
        'for': 'tract:'+tract,
        'in': 'state:'+state+' county:'+county,
        'key': '829bf6f2e037372acbba32ba5731647c5127fdb0'
      }
    # Aggregate City data is best retrieved seperatly rather than as an aggregate of its constituent tracts
    def getCityParams(keys): return {
        'get': 'NAME'+keys,
        'for': 'county:'+county,
        'in': 'state:'+state,
        'key': '829bf6f2e037372acbba32ba5731647c5127fdb0'
      }
    # Called in AddKeys(). Requests data by url and preformats it.
    def readIn( url ):
        tbl = pd.read_json(url, orient='records')
        tbl.columns = tbl.iloc[0]
        return tbl

    # Called by retrieveAcsData.
    # Creates a url and retrieve the data
    # Then appends the city values as tract '010000'
    # Finaly it merges and returns the tract and city totals.
    def addKeys( table, params):
        # Get Tract and City Records For Specific Columns
        table2 = readIn( base+urlencode(getParams(params)) )
        table3 = readIn( base+urlencode(getCityParams(params)) )
        table3['tract'] = '010000'
        # Concatenate the Records
        table2.append([table2, table3], sort=False)
        table2 = pd.concat([table2, table3], ignore_index=True)
        # Merge to Master Table
        table = pd.merge(table, table2,  how='left',
                         left_on=["NAME","state","county","tract"],
                         right_on = ["NAME","state","county","tract"])
        return table

    #~~~~~~~~~~~~~~~
    # Step 1)
    # Retrieve a Meta Data Table Describing the Content of the Table
    #~~~~~~~~~~~~~~~
    url = 'https://api.census.gov/data/20'+year+'/acs/acs5/groups/'+tableId+'.json'
    metaDataTable = pd.read_json(url, orient='records')

    #~~~~~~~~~~~~~~~
    # Step 2)
    # Createa a Dictionary using the Meta Data Table
    #~~~~~~~~~~~~~~~
    # Multiple Queries may be Required.
    # Max columns returned from any given query is 50.
    # For that reasons bin the Columns into Groups of 50.
    for key in metaDataTable['variables'].keys():
      if key[-1:] == 'E':
        keyCount = keyCount + 1
        if keyCount < 40 : keys1 = keys1+','+key
        elif keyCount < 80 : keys2 = keys2+','+key
        elif keyCount < 120 : keys3 = keys3+','+key
        elif keyCount < 160 : keys4 = keys4+','+key
        elif keyCount < 200 : keys5 = keys5+','+key
        elif keyCount < 240 : keys6 = keys6+','+key
        elif keyCount < 280 : keys7 = keys7+','+key
        elif keyCount < 320 : keys8 = keys8+','+key
        keys.append(key)
        val = metaDataTable['variables'][key]['label']
        # Column name formatting
        val = key+'_'+val.replace('Estimate!!', '').replace('!!', '_').replace(' ', '_')
        vals.append(val)
    dictionary = dict(zip(keys, vals))

    #~~~~~~~~~~~~~~~
    # Step 2)
    # Get the actual Table with the data we want using
    #~~~~~~~~~~~~~~~

    # The URL we call is contingent on if the Table we want is a Detailed or Subject table
    url1 = 'https://api.census.gov/data/20'+year+'/acs/acs5?'
    url2 = 'https://api.census.gov/data/20'+year+'/acs/acs5/subject?'
    base = ''
    if tableId[:1] == 'B': base = url1
    if tableId[:1] == 'S': base = url2

    # The addKey function only works after the first set of columns has been downloaded
    # Download First set of Tract columns
    url = base+urlencode(getParams(keys1) )
    table = pd.read_json(url, orient='records')
    table.columns = table.iloc[0]
    table = table.iloc[1:]
    # Download First set of Aggregate City data
    url = base+urlencode(getCityParams(keys1))
    table2 = pd.read_json(url, orient='records')
    table2.columns = table2.iloc[0]
    table2 = table2[1:]
    table2['tract'] = '010000'

    # Merge EM
    #table = pd.concat([table, table2], keys=["NAME","state","county",], axis=0)
    table.append([table, table2], sort=False)
    table = pd.concat([table, table2], ignore_index=True)

    # Now we can repetedly use this function to add as many columns as there are keys listed from the meta data table
    if keys2 != '' : table = addKeys(table, keys2)
    if keys3 != '' : table = addKeys(table, keys3)
    if keys4 != '' : table = addKeys(table, keys4)
    if keys5 != '' : table = addKeys(table, keys5)
    if keys6 != '' : table = addKeys(table, keys6)
    if keys7 != '' : table = addKeys(table, keys7)
    if keys8 != '' : table = addKeys(table, keys8)

    #~~~~~~~~~~~~~~~
    # Step 3)
    # Prepare Column Names using the meta data table. The raw data has columnsNames in the first row, as well.
    # Replace column ID's with labels from the dictionary where applicable (should be always)
    #~~~~~~~~~~~~~~~
    print('Number of Columns', len(dictionary) )

    header = []
    for column in table.columns:
        if column in keys: header.append(dictionary[column])
        else: header.append(column)
    table.columns = header

    # Prettify Names. Only happens with Baltimore...
    table['NAME'] = table['NAME'].str.replace(', Baltimore city, Maryland', '')
    table['NAME'][table['NAME'] == 'Baltimore city, Maryland'] = 'Baltimore City'

    # Convert to Integers Columns from Strings where Applicable
    table = table.apply(pd.to_numeric, errors='ignore')

    # Set the 'NAME' Column as the index dropping the default increment
    table.set_index("NAME", inplace = True)

    if save:

      # Save the raw data as 'TABLEID_5yYEAR.csv'
      table.to_csv('./'+state+county+'_'+tableId+'_5y'+year+'_est_Original.csv', quoting=csv.QUOTE_ALL)

      # Remove the id in the column names & Save the data as 'TABLEID_5yYEAR_est.csv'
      saveThis = table.rename( columns = lambda x : ( str(x)[:] if str(x) in [
        "NAME","state","county","tract"] else str(x)[12:] )  )
      saveThis.to_csv('./'+state+county+'_'+tableId+'_5y'+year+'_est.csv', quoting=csv.QUOTE_ALL)

    return table